import typing
from AcraNetwork.IRIG106.Chapter11 import Chapter11
import struct
import logging

logger = logging.getLogger(__name__)


class FileParser(object):
    """
    Parse a Chapter10 file. Open the file and iterate through it

    >>> fp = FileParser("_dummy.ch10", mode="wb")
    >>> with fp as ch10file:
    ... 	ch10file.write(bytes(10))

    """

    def __init__(self, filename, mode="rb"):
        self.filename = filename
        self._mode = mode
        self.insync = False
        self._offset = 0
        self._fd = None

    def write(self, ch10packet: typing.Union[Chapter11, bytes]) -> None:
        """
        Write a chapter10 packet to the file

        :param ch10packet: The chapter 10 packet to write. Either bytes or Chapter10 object
        """
        if self._fd is None or self._mode != "wb":
            raise Exception("File name not defined")
        if not self._fd.writable():
            raise Exception("File {} not open for writing".format(self.filename))
        if isinstance(ch10packet, Chapter11):
            self._fd.write(ch10packet.pack())
        elif type(ch10packet) is bytes:
            self._fd.write(ch10packet)
        else:
            raise Exception("Write takes a ch10 packet or bytes")

    def __enter__(self):
        self._fd = open(self.filename, self._mode)
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self._fd.close()

    def close(self):
        if self._fd is not None:
            self._fd.close()

    def __iter__(self):
        return self

    def next(self) -> bytes:
        in_sync = False
        pkt_len = 0
        while not in_sync:
            try:
                self._fd.seek(self._offset)
                _first_few_words = self._fd.read(8)
            except:
                raise StopIteration
            try:
                (sync, chid, pkt_len) = struct.unpack("<HHI", _first_few_words)
            except Exception as e:
                logger.debug("Exiting loop err={}".format(e))
                raise StopIteration

            if sync == Chapter11.SYNC_WORD:
                in_sync = True
            else:
                self._offset += 1

        self._fd.seek(self._offset)
        pkt_payload = self._fd.read(pkt_len)
        self._offset += pkt_len
        if len(pkt_payload) != pkt_len:
            raise StopIteration

        return pkt_payload

    __next__ = next



import struct
from AcraNetwork.MPEGTS import MPEGTS
from enum import IntEnum

IPH_OFFSET = 19
TP_OFFSET = 12


class DataStream(IntEnum):
    TRANSPORT = 0x0
    PROGRAM = 0x1


class VideoFormat2(object):
    """
    Chapter 10 Video Format 2
    Object will contain some MPEG Transport streams
    """

    def __init__(self):
        self.channel_specific_word: int = 0
        self.datastream: int = DataStream.TRANSPORT
        self.mpegts: MPEGTS = MPEGTS()

    def unpack(self, buffer: bytes):
        """
        Convert a string buffer into a PCMDataPacket
        :type buffer: bytes
        :rtype: bool
        """

        (self.channel_specific_word,) = struct.unpack_from("<I", buffer, 0)
        iph = bool((self.channel_specific_word >> IPH_OFFSET) & 0x1)
        if iph:
            raise Exception("Option Intra-packet header is not supported")
        self.datastream = DataStream((self.channel_specific_word >> TP_OFFSET) & 0x1)

        self.mpegts.unpack(buffer[4:])

        return True

    def pack(self) -> bytes:
        return struct.pack("<I", self.channel_specific_word) + self.mpegts.pack()

    def __repr__(self):
        _rstr = f"Video Format2={self.channel_specific_word:#0X}\n . MPEGTS={repr(self.mpegts)}"
        return _rstr

    def __eq__(self, other):
        """

        :type other:
        :return:
        """
        if not isinstance(other, VideoFormat2):
            return False

        if self.channel_specific_word != other.channel_specific_word:
            return False

        if self.mpegts != other.mpegts:
            return False

        return True



import struct
from functools import reduce
import logging
from datetime import datetime, timezone
from enum import IntEnum
from decimal import Decimal


DATA_TYPE_TIMEFMT_1 = 0x11
DATA_TYPE_TIMEFMT_2 = 0x12
DATA_TYPE_PCM_DATA_FMT1 = 0x9
DATA_TYPE_MILSTD1553_FMT1 = 0x19
DATA_TYPE_ARINC429_FMT0 = 0x38
DATA_TYPE_UART_FMT0 = 0x50
DATA_TYPE_COMPUTER_GENERATED_FORMAT_0 = 0x0
DATA_TYPE_COMPUTER_GENERATED_FORMAT_1 = 0x1
DATA_TYPE_ANALOG = 0x21
DATA_TYPE_VIDEO_FMT0 = 0x40
DATA_TYPE_VIDEO_FMT1 = 0x41
DATA_TYPE_VIDEO_FMT2 = 0x42
DATA_TYPE_VIDEO_FMT3 = 0x43


class DataType(IntEnum):
    TIMEFORMAT_1 = DATA_TYPE_TIMEFMT_1
    TIMEFORMAT_2 = DATA_TYPE_TIMEFMT_2
    PCM = DATA_TYPE_PCM_DATA_FMT1
    MILSTD1553 = DATA_TYPE_MILSTD1553_FMT1
    ARINC429 = DATA_TYPE_ARINC429_FMT0
    COMPUTER_FORMAT_0 = DATA_TYPE_COMPUTER_GENERATED_FORMAT_0
    COMPUTER_FORMAT_1 = DATA_TYPE_COMPUTER_GENERATED_FORMAT_1
    ANALOG = DATA_TYPE_ANALOG
    UART = DATA_TYPE_UART_FMT0
    VIDEO_FORMAT_0 = DATA_TYPE_VIDEO_FMT0
    VIDEO_FORMAT_1 = DATA_TYPE_VIDEO_FMT1
    VIDEO_FORMAT_2 = DATA_TYPE_VIDEO_FMT2
    VIDEO_FORMAT_3 = DATA_TYPE_VIDEO_FMT3


TS_RTC = 0
TS_SECONDARY = 1

TS_CH4 = 0
TS_IEEE1558 = 1
TS_ERTC = 2


def buf_to_printable(buffer):
    pw = ""
    for idx in range(len(buffer)):
        (v,) = struct.unpack_from(">B", buffer, idx)
        pw += "{:#04X} ".format(v)
        if idx % 8 == 7:
            pw += "\n"
    return pw


class PTPTime(object):
    def __init__(self, seconds=0, nanoseconds=0):
        self.seconds = seconds
        self.nanoseconds = nanoseconds

    def pack(self):
        return struct.pack("<II", self.nanoseconds, self.seconds)

    def unpack(self, buffer):
        (self.nanoseconds, self.seconds) = struct.unpack("<II", buffer)
        return True

    def to_rtc(self):
        ptp_as_date = datetime.fromtimestamp(self.seconds, tz=timezone.utc)
        start_of_year = datetime(ptp_as_date.year, 1, 1, 0, 0, 0)
        start_of_year = start_of_year.replace(tzinfo=timezone.utc)
        seconds_since_start_year = int((ptp_as_date - start_of_year).total_seconds())
        rtc_time = int(seconds_since_start_year * 1e7 + self.nanoseconds / 100)
        return rtc_time

    def to_pinksheet_rtc(self):
        ptp_vector = (Decimal(self.seconds) * Decimal(1e9)) + Decimal(self.nanoseconds)
        ptp_vector_100ns = ptp_vector // Decimal(100)
        ptp_vector_truncated = int(ptp_vector_100ns) & (pow(2, 48) - 1)
        return ptp_vector_truncated

    def __repr__(self):
        time_fmt = "%H:%M:%S %d-%b %Y"
        ts = datetime.fromtimestamp(self.seconds, tz=timezone.utc)
        date_str = ts.strftime(time_fmt)
        return "PTP: {} nanosec={}".format(date_str, self.nanoseconds)

    def __eq__(self, __value):
        if not isinstance(__value, PTPTime):
            return False
        if self.nanoseconds != __value.nanoseconds or self.seconds != __value.seconds:
            return False
        return True

    def __add__(self, val):
        if isinstance(val, PTPTime):
            addns = self.nanoseconds + val.nanoseconds
            ns = int(addns % 1e9)
            sec = self.seconds + val.seconds + int(addns // 1e9)
            return PTPTime(sec, ns)
        else:
            raise Exception("Addition of other PTPTime instances")

    def __sub__(self, val):
        if isinstance(val, PTPTime):
            if val.nanoseconds > self.nanoseconds:
                sec = self.seconds - val.seconds - 1
                nsec = int(1e9) - (val.nanoseconds - self.nanoseconds)
            else:
                sec = self.seconds - val.seconds
                nsec = self.nanoseconds - val.nanoseconds
            return PTPTime(sec, nsec)
        else:
            raise Exception("Subtraction of other PTPTime instances")

    def __lt__(self, val):
        if isinstance(val, PTPTime):
            return (self.seconds + self.nanoseconds / 1e9) < (val.seconds + val.nanoseconds / 1e9)

    def __le__(self, val):
        if isinstance(val, PTPTime):
            _d = (self.seconds + self.nanoseconds / 1e9) <= (val.seconds + val.nanoseconds / 1e9)
            return _d


class RTCTime(object):
    def __init__(self, count=0):
        self.count = count

    def pack(self):
        msw = (self.count >> 32) & 0xFFFF
        lsw = self.count & 0xFFFFFFFF
        return struct.pack("<IHH", lsw, msw, 0)

    def unpack(self, buffer):
        (lsw, msw, _zero) = struct.unpack("<IHH", buffer)
        self.count = lsw + (msw << 32)
        return True

    def to_rtc(self):
        return self.count

    def to_pinksheet_rtc(self):
        return self.count

    def __repr__(self):
        return "RTC: count={}".format(self.count)

    def __eq__(self, __value):
        if not isinstance(__value, RTCTime):
            return False
        if self.count != __value.count:
            return False
        return True


logger = logging.getLogger(__name__)


def get_checksum_buf(buf: bytes) -> int:
    """
    Return the arithmetic checksum of a header. This is the checksum that is used for the chapter 10 header

    :param buf:
    :return:
    """
    if len(buf) % 2 != 0:
        raise Exception("buffer needs to be 16-bit aligned")

    words = struct.unpack("<{}H".format(len(buf) // 2), buf)
    sum = reduce(lambda x, y: x + y, words)

    return sum % 65536


def get_checksum_byte_buf(buf: bytes) -> int:
    """
    Return the arithmetic checksum of a header. This is the checksum used for the secondary header checksu,

    :param buf:
    :return:
    """

    checksumbytes = struct.unpack(f"<{len(buf)}B", buf)
    sum = reduce(lambda x, y: x + y, checksumbytes)

    return sum % 65536


class Chapter11(object):
    """
    Class to pack and unpack Chapter11 packets. These can not be directly transmitted over a network
    but they can be written to .ch10 files

    Create a packet and transmit it via UDP

    >>> from AcraNetwork.IRIG106.Chapter11 import DataType
    >>> from AcraNetwork.IRIG106.Chapter10.FileParser import FileParser
    >>> import AcraNetwork.IRIG106.Chapter11.ComputerData as chcomputer
    >>> fp = FileParser("myfile.ch10", mode="wb")
    >>> c = Chapter11()
    >>> c.channelID = 0
    >>> c.sequence = 0
    >>> c.packetflag = 0
    >>> c.datatype = DataType.COMPUTER_FORMAT_1
    >>> c.relativetimecounter = 0
    >>> ctmats = chcomputer.ComputerGeneratedFormat1()
    >>> ctmats.payload = bytes(3)
    >>> c.payload = ctmats.pack()
    >>> with fp as ch10file:
    ...     ch10file.write(c)


    """

    SYNC_WORD = 0xEB25  # :(Object Constant) Sync word

    CH10_HDR_FORMAT = "<HHIIBBBBIHH"
    CH10_HDR_FORMAT_LEN = struct.calcsize(CH10_HDR_FORMAT)

    CH10_OPT_HDR_FORMAT = "<IIHH"
    CH10_OPT_HDR_FORMAT_LEN = struct.calcsize(CH10_OPT_HDR_FORMAT)

    TS_SOURCES = [TS_RTC, TS_SECONDARY]  # :(Object Constant) Valid timesources, assign to :attr:`Chapter10.ts_source`
    TS_SECONDARY_SOURCES = [TS_CH4, TS_IEEE1558, TS_ERTC]

    PKT_FLAG_SECONDARY = 0x80  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_SEC_HDR_TIME = 0x40  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RTC_SYNC_ERROR = 0x20  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_DATA_OVERFLOW = 0x10  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RCC_TIME = 0x0  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_1588_TIME = 0x04  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_8BIT_CHKSUM = 0x1  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_16BIT_CHKSUM = 0x2  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_32BIT_CHKSUM = 0x3  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable

    def __init__(self):
        self.syncpattern = (
            Chapter11.SYNC_WORD
        )  # :(2 Bytes) contains a static sync value for the every packet. The Packet Sync Pattern value shall be 0xEB25
        self.channelID = 0  # :(2 Bytes) contains a value representing the Packet Channel ID.
        self.packetlen = 0  # :(4 Bytes) contains a value representing the length of the entire packet. The value shall be in bytes and is always a multiple of four
        self.datalen = 0  # :(4 Bytes) contains a value representing the valid data length within the packet
        self.datatypeversion = 0x5  #: RCC released versions
        self.sequence = 0  # :(1 Byte) contains a value representing the packet sequence number for each Channel ID.
        self._packetflag = (
            0  # :(1 Byte) contains bits representing information on the content and format of the packet(s)
        )
        self.datatype = 0  # :(1 Byte) contains a value representing the type and format of the data
        self.relativetimecounter = 0  # :(6 Bytes) contains a value representing the 10 MHz Relative Time Counter (RTC)
        self.ptptime = PTPTime()  #: PTP Timestamp
        self.ts_source = TS_RTC  # :The timestamp source. Select from :attr:`Chapter10.TS_SOURCES`
        self.payload = b""  # :The payload
        self.data_checksum_size = 0
        self.filler = b""
        self.has_secondary_header = False

    @property
    def packetflag(self):
        """
        (1 Byte) contains bits representing information on the content and format of the packet(s)

        :type val: intx
        """
        return self._packetflag

    @packetflag.setter
    def packetflag(self, val):
        if val > 0xFF:
            raise Exception("Packet flag ={:#0X} to valud".format(val))
        self._packetflag = val

        if self._packetflag >> 7 == 1:
            if ((self._packetflag >> 2) & 0x3) == 0:
                self.ts_source = TS_CH4
            elif ((self._packetflag >> 2) & 0x3) == 1:
                self.ts_source = TS_IEEE1558
            else:
                raise Exception("Time format is illegal")

            self.has_secondary_header = True
        else:
            self.has_secondary_header = False
            self.ts_source = TS_RTC

    def pack(self):
        """
        Pack the Chapter10 object into a binary buffer

        :rtype: bytes
        """

        if self.has_secondary_header:
            if self.ts_source == TS_CH4:
                raise Exception("Ch4 Timestamp in secondary header not supported")

            sec_hdr = self.ptptime.pack() + struct.pack(">HH", 0, 0)
            # Replace the checksum
            cs = get_checksum_byte_buf(sec_hdr)
            sec_hdr = sec_hdr[:-2] + struct.pack("<H", cs)
        else:
            sec_hdr = b""

        total_len_excl_filler = (
            len(sec_hdr) + len(self.payload) + Chapter11.CH10_HDR_FORMAT_LEN + self.data_checksum_size
        )

        # Add the filler
        if total_len_excl_filler % 4 == 0:
            self.filler = b""
        else:
            fill_len = 4 - (total_len_excl_filler % 4)
            fill_val = [0xFF] * fill_len
            self.filler = struct.pack(">{}B".format(fill_len), *fill_val)

        self.packetlen = total_len_excl_filler + len(self.filler)
        self.datalen = len(self.payload)

        checksum = 0
        _rtc_lwr = self.relativetimecounter & 0xFFFFFFFF
        _rtc_upr = self.relativetimecounter >> 32

        hdr = struct.pack(
            Chapter11.CH10_HDR_FORMAT,
            self.syncpattern,
            self.channelID,
            self.packetlen,
            self.datalen,
            self.datatypeversion,
            self.sequence,
            self.packetflag,
            self.datatype,
            _rtc_lwr,
            _rtc_upr,
            checksum,
        )
        hdr = hdr[:-2] + struct.pack("<H", get_checksum_buf(hdr))
        return hdr + sec_hdr + self.payload + self.filler

    def unpack(self, buffer):
        """
        Unpack a string buffer into an Chapter10 object

        :param buffer: A string buffer representing an Chapter10 packet
        :type buffer: bytes
        :rtype: None
        """
        (
            self.syncpattern,
            self.channelID,
            self.packetlen,
            self.datalen,
            self.datatypeversion,
            self.sequence,
            self.packetflag,
            self.datatype,
            _rtc_lwr,
            _rtc_upr,
            checksum,
        ) = struct.unpack_from(Chapter11.CH10_HDR_FORMAT, buffer)

        offset_hdr = struct.calcsize(Chapter11.CH10_HDR_FORMAT) - 2
        exp_checksum = get_checksum_buf(buffer[:offset_hdr])
        if checksum != exp_checksum:
            logger.error("Ch10 Header checksum {:#0X} does not match expected={:#0X}".format(checksum, exp_checksum))

        self.relativetimecounter = _rtc_lwr + (_rtc_upr << 32)

        if (self.packetflag >> 7) == 1:
            self.has_secondary_header = True
            pkt_hdr_time = (self.packetflag >> 2) & 0x3
            (ts_ns, ts_s, _res, _checksum_sec) = struct.unpack_from(
                Chapter11.CH10_OPT_HDR_FORMAT, buffer, Chapter11.CH10_HDR_FORMAT_LEN
            )

            sec_exp_checksum = get_checksum_byte_buf(
                buffer[
                    Chapter11.CH10_HDR_FORMAT_LEN : Chapter11.CH10_HDR_FORMAT_LEN
                    + Chapter11.CH10_OPT_HDR_FORMAT_LEN
                    - 2
                ]
            )

            if _checksum_sec != sec_exp_checksum:
                logger.error(
                    "Ch10 Secondary Header checksum ({:#0X}) does not match expected={:#0X}".format(
                        _checksum_sec, sec_exp_checksum
                    )
                )
                # print("Ch10 Secondary Header checksum does not match expected={:#0X}".format(sec_exp_checksum))

            if pkt_hdr_time == 0:
                raise Exception("Ch4 Timestamp in secondary header not supported")
                # print("Ch4 Timestamp in secondary header not supported")
            elif pkt_hdr_time == 1:
                self.ts_source = TS_IEEE1558
                self.ptptime = PTPTime(ts_s, ts_ns)
            else:
                raise Exception("Secondary Header Time Format not legal")
                # print("Secondary Header Time Format not legal")
            self.payload = buffer[(Chapter11.CH10_HDR_FORMAT_LEN + Chapter11.CH10_OPT_HDR_FORMAT_LEN) :]
        else:
            self.has_secondary_header = False
            self.payload = buffer[Chapter11.CH10_HDR_FORMAT_LEN :]
            self.ts_source = TS_RTC

        return True

    def __eq__(self, other):
        if not isinstance(other, Chapter11):
            return False

        _match_att = (
            "syncpattern",
            "channelID",
            "packetlen",
            "datalen",
            "datatypeversion",
            "datatype",
            "_packetflag",
            "relativetimecounter",
            "ptptime",
            "ts_source",
            "payload",
            "data_checksum_size",
            "filler",
            "has_secondary_header",
        )

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "Chapter 11: ChannelID={} Sequence={} DataLen={}".format(self.channelID, self.sequence, self.datalen)



from datetime import datetime, timezone
import struct
from AcraNetwork.IRIG106.Chapter11 import PTPTime

ITS_FREEWHEELING = 0x0 << 12
ITS_FREEWHEELING_FROM_TIME = 0x0 << 12
ITS_FREEWHEELING_FROM_RMM = 0x0 << 12
ITS_LOCKED_IRIG = 0x0 << 12
ITS_LOCKED_GPS = 0x0 << 12
ITS_LOCKED_NTP = 0x0 << 12
ITS_LOCKED_PTP = 0x0 << 12
ITS_LOCKED_EMBEDDED = 0x0 << 12

DATE_FMT_YEAR_AVAIL = 0x1 << 9
DATE_FMT_LEAP_YEAR = 0x1 << 8

TIME_FMT_IRIGB = 0x0
TIME_FMT_IRIGA = 0x1 << 4
TIME_FMT_IRIGG = 0x2 << 4
TIME_FMT_RTC = 0x3 << 4
TIME_FMT_UTC = 0x4 << 4
TIME_FMT_GPS = 0x5 << 4
TIME_FMT_NONE = 0xF << 4

SRC_INTERNAL = 0x0
SRC_EXTERNAL = 0x1
SRC_INT_FMM = 0x2
SRC_NONE = 0xF


def double_digits_to_bcd(val: int) -> int:
    """
    Very simplified conversion of time to format used in Ch10 Time packet

    :param val: time field value in integers.
    :return:
    """
    offsets = {1: 0, 10: 4}
    retval = 0
    for dec, offset in offsets.items():
        retval += (int(val / dec) % 10) << offset
    return retval


def bcd_to_int(val: int) -> int:
    """

    :param val:
    :return:
    """
    if val < 0:
        raise ValueError("Cannot be a negative integer")

    if val == 0:
        return 0

    bcdstring = ""
    while val > 0:
        nibble = val % 16
        bcdstring = str(nibble) + bcdstring
        val >>= 4
    return int(bcdstring)


class TimeDataFormat1(object):
    """
    Class to pack and unpack Chapter10 TIME packet payloads.

    Create a packet and transmit it via UDP

    >>> from base64 import b64encode
    >>> t = TimeDataFormat1()
    >>> t.ptptime.seconds = 100
    >>> print(len(t.pack()))
    12

    :type channel_specific_data: int
    :type milliseconds: int
    :type datetime: datetime | None
    :type filler: str
    """

    def __init__(self):
        self.channel_specific_data = ITS_FREEWHEELING + DATE_FMT_YEAR_AVAIL + TIME_FMT_GPS + SRC_EXTERNAL
        self.ptptime = PTPTime(0, 0)

    def pack(self):
        """
        Pack the Chapter10 object into a binary buffer

        :rtype: str
        """

        dt = datetime.fromtimestamp(self.ptptime.seconds, tz=timezone.utc)
        packet_bytes = list()
        packet_bytes.append(double_digits_to_bcd(self.ptptime.nanoseconds / 1e7))
        packet_bytes.append(double_digits_to_bcd(dt.second))
        packet_bytes.append(double_digits_to_bcd(dt.minute))
        packet_bytes.append(double_digits_to_bcd(dt.hour))
        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            packet_bytes.append(double_digits_to_bcd(dt.day))
            packet_bytes.append(double_digits_to_bcd(dt.month))
            packet_bytes.append(double_digits_to_bcd(dt.year % 100))
            packet_bytes.append(double_digits_to_bcd(dt.year / 100))
        else:
            doy = int(dt.strftime("%j"))
            packet_bytes.append(double_digits_to_bcd(doy % 100))
            packet_bytes.append(double_digits_to_bcd(doy / 100))

        mybytes = struct.pack("<I", self.channel_specific_data)
        mybytes += struct.pack("<{}B".format(len(packet_bytes)), *packet_bytes)

        return mybytes

    def unpack(self, buf):
        """
        Unpack a string buffer into a TimeFormat object
        :param buffer:
        :return:
        """
        (self.channel_specific_data, ms, s, mn, h) = struct.unpack_from("<IBBBB", buf)
        milliseconds = bcd_to_int(ms) * 10
        nanoseconds = milliseconds * int(1e6)
        sec = bcd_to_int(s)
        mins = bcd_to_int(mn)
        hrs = bcd_to_int(h)

        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            (d, mo, yr) = struct.unpack_from("<BBH", buf, 8)
            day = bcd_to_int(d)
            mon = bcd_to_int(mo)
            year = bcd_to_int(yr)
            dt = datetime(year, mon, day, hrs, mins, sec)
        else:
            (doy, hdoy) = struct.unpack_from("<BB", buf, 8)
            day_of_year = bcd_to_int(doy) + 100 * bcd_to_int(hdoy)
            date_as_string = "{:02d}:{:02d}:{:02d} {:03d} 1970 GMT".format(hrs, mins, sec, day_of_year)
            dt = datetime.strptime(date_as_string, "%H:%M:%S %j %Y %Z")

        dt = dt.replace(tzinfo=timezone.utc)

        seconds = int(dt.timestamp())
        self.ptptime = PTPTime(seconds, nanoseconds)

        return True

    def __eq__(self, other):
        if not isinstance(other, TimeDataFormat1):
            return False

        _match_att = ("channel_specific_data", "ptptime")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "TimeFormat1 ChannelSpecificWord={:#0X} Time={}".format(self.channel_specific_data, self.ptptime)

    def __len__(self):
        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            return 4 * (4 + 1)
        else:
            return 4 * (3 + 1)


TS_STATUS_VALID = 0x1
TS_STATUS_IEEE2002 = 0x1 << 4
TS_STATUS_IEEE2008 = 0x2 << 4


class TimeDataFormat2(object):
    """
    Class to pack and unpack Chapter10 TIME packet payloads.

    Create a packet and transmit it via UDP


    :type channel_specific_data: int
    :type filler: str
    """

    def __init__(self):
        self.channel_specific_data = TS_STATUS_VALID + TS_STATUS_IEEE2002
        self.ptptime = PTPTime()

    def pack(self):
        """
        Pack the Chapter10 time object into a binary buffer

        :rtype: str
        """

        if (self.channel_specific_data >> 4) & 0x1:
            # ptp
            frac_sec = self.ptptime.nanoseconds
        else:
            # ntp
            frac_sec = int(self.ptptime.nanoseconds * (pow(2, 32) / 1e9))

        return struct.pack("<III", self.channel_specific_data, self.ptptime.seconds, frac_sec)

    def unpack(self, buf):
        """
        Unpack a string buffer into a TimeFormat object
        :param buffer: str
        :return:
        """
        (self.channel_specific_data, s, fs) = struct.unpack("<III", buf)
        if (self.channel_specific_data >> 4) & 0xF != 0:  # Make PTP the default
            nanoseconds = fs
        else:
            nanoseconds = int(fs / (pow(2, 32) / 1e9))  # NTP

        self.ptptime = PTPTime(s, nanoseconds)

        return True

    def __eq__(self, other):
        if not isinstance(other, TimeDataFormat2):
            return False

        _match_att = ("channel_specific_data", "ptptime")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "TimeFormat2 ChannelSpecificWord={:#0X} Time={} ".format(self.channel_specific_data, self.ptptime)

    def __len__(self):
        return 3 * 4



from __future__ import annotations

"""
.. module:: SimpleEthernet
    :platform: Unix, Windows
    :synopsis: A very trimmed down set of classes to unpack the common network packet formats

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import socket
from functools import reduce
from zlib import crc32
import logging
import typing
import enum


logger = logging.getLogger(__name__)


def unpack48(x: bytes) -> int:
    """
    Unpack a 48bit string returning an integer

    :param x: 6 byte buffer
    :type x: bytes

    :rtype: int
    """
    x2, x3 = struct.unpack(">HI", x)
    return x3 | (x2 << 32)


def pack48(x: int) -> bytes:
    return struct.pack(">HI", x >> 32, x & 0xFFFF_FFFF)


def mactoreadable(macaddress: int) -> str:
    """
    Convert a macaddress into the readable form

    :param macaddress: The mac address in integer format
    :type macaddress: int

    :rtype: str
    """
    mac_string = ""
    b = []
    for i in range(6):
        eachbyte = (macaddress >> i * 8) & 0xFF
        b.append(eachbyte)

    return "{:02X}:{:02X}:{:02X}:{:02X}:{:02X}:{:02X}".format(b[5], b[4], b[3], b[2], b[1], b[0])


def ip_calc_checksum(pkt: bytes) -> int:
    """
    Calculate the checksum of a packet

    :param pkt: The IP packet header packed into bytes
    :type pkt: str|bytes
    :return:
    """

    if len(pkt) % 2 == 1:
        pkt += b"\0"
    s = sum(struct.unpack("<{}H".format(len(pkt) // 2), pkt))
    s = (s >> 16) + (s & 0xFFFF)
    s += s >> 16
    s = ~s
    return s & 0xFFFF


class EthType(enum.IntEnum):
    TYPE_IP = 0x800
    TYPE_IPv4 = 0x800
    TYPE_IPv6 = 0x86DD
    TYPE_ARP = 0x806
    TYPE_VLAN = 0x8100
    TYPE_PAUSE = 0x8808


class Ethernet(object):
    """
    This is simple class to pack or unpack an Ethernet packet. Handles very basic packets that are used in FTI

    Read an Ethernet Packet from a pcap file

    >>> from base64 import b64decode
    >>> raw_packet = b64decode('AQBeAAABAAxNAApsCABFAAA61p1AAP8R3VrAqBwQ6wAAAQP/H0oAJgAAAB8ADwAPA1ffwH8A1pwAAQvUQGAAAP3NEAEoHP//')
    >>> e = Ethernet()
    >>> e.unpack(raw_packet)
    True
    >>> print(e)
    SRCMAC=00:0C:4D:00:0A:6C DSTMAC=01:00:5E:00:00:01 TYPE=0X800

    """

    HEADERLEN = 14
    HEADERLEN_VLAN = 18
    TYPE_IP = EthType.TYPE_IPv4
    TYPE_IPv4 = EthType.TYPE_IPv4  # :(Object Constant) IPv4 Type Constant
    TYPE_IPv6 = EthType.TYPE_IPv6  # :(Object Constant) IPv6 Type Constant
    TYPE_ARP = EthType.TYPE_ARP  # :(Object Constant) ARP Type Constant
    TYPE_PAUSE = EthType.TYPE_PAUSE  # :(Object Constant) PAUSE Type Constant
    TYPE_VLAN = EthType.TYPE_VLAN  # :(Object Constant) VLAN Type Constant
    ADDR_LENGTH = 6  # Address length for ARP

    def __init__(self, buf: typing.Optional[bytes] = None):
        """
        Create an Ethernet packet object.

        :param buf: If a buffer is passed in to the init method, it will be unpacked as a Ethernet packet
        :type buf: bytes

        """
        self.type: int = (
            Ethernet.TYPE_IP
        )  #: The Ethertype field. Assign using the TYPE_* constants. https://en.wikipedia.org/wiki/EtherType
        self.srcmac: int = (
            0x0  #: The Ethernet source MAC Address. This is encoded into a 48bit field. https://en.wikipedia.org/wiki/MAC_address
        )
        self.dstmac: int = (
            0x0  #: The Ethernet destination MAC Address. This is encoded into a 48bit field. https://en.wikipedia.org/wiki/MAC_address
        )
        self.payload: bytes = bytes()  #: The Ethernet payload. Typically an IP packet.
        self.vlan: bool = False
        self.vlantag: int = 0xFFFF

        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf: bytes, fcs=False) -> bool:
        """
        Unpack a raw byte stream to an Ethernet object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :param fcs: Assume FCS is included in the buffer
        :type fcs: bool
        :rtype: bool
        """

        self.dstmac = unpack48(buf[:6])
        self.srcmac = unpack48(buf[6:12])
        (_type,) = struct.unpack_from(">H", buf, 12)
        if _type == Ethernet.TYPE_VLAN:
            self.vlan = True
            (self.vlantag, self.type) = struct.unpack_from(">HH", buf, 14)
            hdr_len = Ethernet.HEADERLEN_VLAN
        else:
            self.type = _type
            hdr_len = Ethernet.HEADERLEN
        if fcs:
            self.payload = buf[hdr_len:-4]
            _fcs_buf = buf[-4:]
            exp_crc = crc32(buf[:-4]) & 0xFFFFFFFF
            (act_crc,) = struct.unpack("I", _fcs_buf)
            if exp_crc != act_crc:
                raise Exception("FCS is wrong. Exo={:#0X} Act={:#0X}".format(exp_crc, act_crc))
        else:
            self.payload = buf[hdr_len:]
        return True

    def pack(self, fcs: bool = False) -> bytes:
        """
        Pack the Ethernet object into a buffer

        :param fcs: Include FCS in the buffer returned
        :rtype: bytes
        """
        if self.dstmac is None or self.srcmac is None or self.type is None or self.payload is None:
            raise ValueError("All three required Ethernet fields are not complete")
        if self.vlan:
            header = struct.pack(
                ">HIHIHHH",
                self.dstmac >> 32,
                self.dstmac & 0xFFFFFFFF,
                self.srcmac >> 32,
                self.srcmac & 0xFFFFFFFF,
                Ethernet.TYPE_VLAN,
                self.vlantag,
                self.type,
            )
        else:
            header = struct.pack(
                ">HIHIH",
                self.dstmac >> 32,
                self.dstmac & 0xFFFFFFFF,
                self.srcmac >> 32,
                self.srcmac & 0xFFFFFFFF,
                self.type,
            )

        if fcs:
            _crc = crc32(header + self.payload) & 0xFFFFFFFF
            return header + self.payload + struct.pack("I", _crc)
        else:
            return header + self.payload

    def __repr__(self):
        if self.vlan:
            return "SRCMAC={} DSTMAC={} TYPE={:#0X} VLAN={}".format(
                mactoreadable(self.srcmac), mactoreadable(self.dstmac), self.type, self.vlantag
            )
        else:
            return "SRCMAC={} DSTMAC={} TYPE={:#0X}".format(
                mactoreadable(self.srcmac), mactoreadable(self.dstmac), self.type
            )

    def __eq__(self, other):
        if not isinstance(other, Ethernet):
            return False

        for attr in ["type", "dstmac", "srcmac", "payload", "vlan", "vlantag"]:
            if getattr(self, attr) != getattr(other, attr):
                # print("Attribute {} does not match {}/{}".format(attr,  getattr(self, attr), getattr(other, attr) ))
                return False

        return True


class IP(object):
    """
    Create or unpack an IP packet https://en.wikipedia.org/wiki/IPv4#Header

    If you wanted to unpack an Ethernet object payload which contains an IP packet

    >>> from base64 import b64decode
    >>> raw_packet = b64decode('AQBeAAABAAxNAApsCABFAAA61p1AAP8R3VrAqBwQ6wAAAQP/H0oAJgAAAB8ADwAPA1ffwH8A1pwAAQvUQGAAAP3NEAEoHP//')
    >>> i = IP()
    >>> i.unpack(raw_packet[0x10:])
    True

    """

    PROTOCOL_ICMP = 0x01  # :(Object Constant) ICMP Protocol Constant
    PROTOCOL_IGMP = 0x02  # :(Object Constant) IGMP Protocol Constant
    PROTOCOL_TCP = 0x6  # :(Object Constant) TCP Protocol Constant
    PROTOCOL_UDP = 0x11  # :(Object Constant) UDP Protocol Constant
    FLAG_DONT_FRAGMENT = 0x2
    FLAG_MORE_FRAGMENTS = 0x1

    PROTOCOLS = {
        "ICMP": PROTOCOL_ICMP,
        "IGMP": PROTOCOL_IGMP,
        "TCP": PROTOCOL_TCP,
        "UDP": PROTOCOL_UDP,
    }  # (Object Constant) Protocols available
    IP_HEADER_FORMAT = ">BBHHBBBBHII"
    IP_HEADER_SIZE = struct.calcsize(IP_HEADER_FORMAT)
    ADDR_LENGTH = 4

    def __init__(self, buf: typing.Optional[bytes] = None):
        """
        Create an IP packet object. Currently supports only IPv4

        :param buf: If a buffer is passed in to the init method, it will be unpacked as a IP packet
        :type buf: bytes

        """
        self.srcip: str = ""  #: Source IP Address
        self.dstip: str = ""  #: Destination IP Address
        self.len: int = 0  #: Total Length. This is calculated when packing the packet
        self.flags: int = 0x0  #: Three bit field identifying a flag
        self.fragment_offset: int = 0  #: Fragment offset
        self.protocol: int = IP.PROTOCOL_UDP  #: The type of the payload
        self.payload: bytes = bytes()  #: The IPv4 payload
        self.version: int = 4  #: IP version field
        self.ihl: int = 5  #: Header length in 32 bit words
        self.dscp: int = 0  #: Differentiated Services Code Point
        self.id: int = 0  #: Identification Field
        self.ttl: int = 20  #: Time to Live. In practice the hop count.

        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf: bytes) -> bool:
        """
        Unpack a raw byte stream to an IP object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """
        if len(buf) < IP.IP_HEADER_SIZE:
            raise ValueError("Buffer too short for to be an IP packet")
        (
            na1,
            self.dscp,
            self.len,
            self.id,
            self.flags,
            na3,
            self.ttl,
            self.protocol,
            checksum,
            self.srcip,
            self.dstip,
        ) = struct.unpack_from(IP.IP_HEADER_FORMAT, buf)
        self.fragment_offset = (((self.flags & 0x1F) << 8) + na3) * 8
        self.flags = self.flags >> 5
        self.version = na1 >> 4
        self.ihl = na1 & 0xF
        self.srcip = socket.inet_ntoa(struct.pack("!I", self.srcip))
        self.dstip = socket.inet_ntoa(struct.pack("!I", self.dstip))
        # Fill IP payload with number of bytes declared in header's length field, leaving any trailer behind (e.g. typically padding to reach 64bytes)
        self.payload = buf[IP.IP_HEADER_SIZE : self.len]

        computed_ip_checksum = ip_calc_checksum(buf[: IP.IP_HEADER_SIZE])
        if computed_ip_checksum != 0:
            logger.error(
                f"Invalid IP Header Checksum. Computed Checksum = 0x{computed_ip_checksum:04X} (should be 0x0000). Raw received Checksum = 0x{checksum:04X}."
            )

        return True

    def pack(self) -> bytes:
        """
        Pack the IP object into a buffer

        :rtype: bytes
        """

        for word in [self.dscp, self.id, self.flags, self.ttl, self.protocol, self.srcip, self.dstip]:
            if word is None:
                raise ValueError(f"All required IP payloads not defined. Attribute {word} is None")

        (srcip_as_int,) = struct.unpack("!I", socket.inet_aton(self.srcip))
        (dstip_as_int,) = struct.unpack("!I", socket.inet_aton(self.dstip))
        self.len = IP.IP_HEADER_SIZE + len(self.payload)
        if self.len > 65536:
            logger.warning("IP Payload longer than 65536. Truncating the length field")
        header = struct.pack(
            IP.IP_HEADER_FORMAT,
            0x45,
            self.dscp,
            self.len % 65536,
            self.id,
            self.flags,
            0,
            self.ttl,
            self.protocol,
            0,
            srcip_as_int,
            dstip_as_int,
        )
        checksum = ip_calc_checksum(header)
        header = header[:10] + struct.pack("H", checksum) + header[12:]
        return header + self.payload

    def __repr__(self):
        protocol = ""
        for p, v in IP.PROTOCOLS.items():
            if v == self.protocol:
                protocol = p
        return "SRCIP={} DSTIP={} PROTOCOL={} LEN={}".format(self.srcip, self.dstip, protocol, self.len)


class IPv4(IP):
    pass


class IPv6(object):
    """
    Create an IPv6 packet https://en.wikipedia.org/wiki/IPv6_packet

    :type version
    :type traffic_class
    :type flow_label
    :type len
    :type next_header
    :type hop_limit
    :type srcip
    :type dstip

    """

    FLAG_DONT_FRAGMENT = 0x2
    FLAG_MORE_FRAGMENTS = 0x1

    IP_HEADER_FORMAT = ">IHBBIIIIIIII"
    IP_HEADER_SIZE = struct.calcsize(IP_HEADER_FORMAT)

    def __init__(self, buf=None):
        """

        :param buf: If a buffer is passed in to the init method, it will be unpacked as a IP packet
        :type buf: bytes

        """
        self.version: int = 0x6  #: IP version field
        self.traffic_class: int = 0x00
        self.flow_label: int = 0x00000
        self.len: int = 0  #: Total Length. This is calculated when packing the packet
        self.next_header: int = 0x3B  # 0x3B = No next header
        self.hop_limit: int = 0x01
        self.srcip: int = 0  #: Source IP Address
        self.dstip: int = 0  #: Destination IP Address
        self.payload: bytes = bytes()

        if buf is not None:
            self.unpack(buf)

    def pack(self) -> bytes:
        """
        Pack the IP object into a buffer

        :rtype: bytes
        """

        for word in [self.srcip, self.dstip]:
            if word is None:
                raise ValueError("All required IP payloads not defined")

        self.len = len(self.payload)
        header = struct.pack(
            IPv6.IP_HEADER_FORMAT,
            ((self.version << 28) + (self.traffic_class << 24) + self.flow_label),
            self.len,
            self.next_header,
            self.hop_limit,
            self.srcip >> 96,
            (self.srcip >> 64) & 0xFFFFFFFF,
            (self.srcip >> 32) & 0xFFFFFFFF,
            self.srcip & 0xFFFFFFFF,
            self.dstip >> 96,
            (self.dstip >> 64) & 0xFFFFFFFF,
            (self.dstip >> 32) & 0xFFFFFFFF,
            self.dstip & 0xFFFFFFFF,
        )

        return header + self.payload

    def unpack(self, buffer):
        raise Exception("Not implemented")

    def __repr__(self):
        return "SRCIP={} DSTIP={} LEN={}".format(self.srcip, self.dstip, self.len)


class UDP(object):
    """
    Class to build and unpack a UDP packet

    https://en.wikipedia.org/wiki/User_Datagram_Protocol#Packet_structure

    Packet structure::

        -----2B----- -----2B----- -----2B----- -----2B----- --0-65527B----
        | SRC PORT  |  DEST PORT |   LENGTH   | CHECKSUM   | PAYLOAD
        ------------ ------------ ------------ ------------ --------------

    Create a UDP packet

    >>> u = UDP()
    >>> u.dstport = 5500
    >>> u.srcport = 4400
    >>> u.payload = struct.pack('B',0x5)
    >>> mypacket = u.pack()


    """

    UDP_HEADER_FORMAT = ">HHHH"
    UDP_HEADER_SIZE = struct.calcsize(UDP_HEADER_FORMAT)

    def __init__(self, buf: typing.Optional[bytes] = None):
        self.srcport: int = 0  #: The UDP source port number
        self.dstport: int = 0  #: The UDP desitnation port number
        self.len: int = 0  #: The length of the UDP header and payload in bytes
        self.payload: bytes = bytes()  #: The UDP payload

        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf: bytes):
        """
        Unpack a raw byte stream to a UDP object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        if len(buf) < UDP.UDP_HEADER_SIZE:
            raise ValueError("Buffer too short to be a UDP packet")
        (self.srcport, self.dstport, self.len, checksum) = struct.unpack_from(UDP.UDP_HEADER_FORMAT, buf)
        self.payload = buf[UDP.UDP_HEADER_SIZE :]

        return True

    def pack(self) -> bytes:
        """
        Pack the UDP object into a buffer

        :rtype: bytes
        """

        if self.srcport is None or self.dstport is None or self.payload is None:
            raise ValueError("All UDP fields need to be defined to pack the payload")

        self.len = len(self.payload) + UDP.UDP_HEADER_SIZE
        if self.len >= 65536:
            logger.warning("UDP Payload longer than 65536. Truncating the length field")
        return struct.pack(UDP.UDP_HEADER_FORMAT, self.srcport, self.dstport, self.len % 65536, 0) + self.payload

    def __repr__(self):
        return "SRCPORT={} DSTPORT={}".format(self.srcport, self.dstport)


class AFDX(object):
    """
    This class will  unpack an AFDX packet

    """

    HEADERLEN = 14
    DSTMAC_CONST = 0x3000000
    SRCMAC_CONST = 0x20000
    MIN_PAYLOAD_LEN = 42

    def __init__(self, buf=None):
        raise Exception("No working")
        self.type = None
        self.networkID = None
        self.equipmentID = 0
        self.interfaceID = 0
        self.vlink = None

        self.payload = None
        self.sequencenum = None
        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf):
        self.set_dstmac(buf[:6])
        self.unpacksrcmac(unpack48(buf[6:12]))

        (self.type,) = struct.unpack_from(">H", buf, 12)
        self.payload = buf[AFDX.HEADERLEN : -1]
        self.sequencenum = struct.unpack("B", buf[-1])

    def unpacksrcmac(self, mac):
        srcconstantf = mac >> 24
        # if srcconstantf != AFDX.SRCMAC_CONST:
        #    raise ValueError('Expected constant field of {:#x} in SrcMac Address'.format(AFDX.SRCMAC_CONST))
        # (self.networkID,self.equipmentID,self.interfaceID) = struct.unpack_from('BBB',mac[:3])
        # self.interfaceID = self.interfaceID >> 5

    def set_dstmac(self, mac):
        (dstconstantf, vlink) = struct.unpack_from(">IH", mac)
        # if dstconstantf != AFDX.DSTMAC_CONST:
        #    raise ValueError('Expected constant field of {:#x} in DestMac Address'.format(AFDX.DSTMAC_CONST))
        self.vlink = vlink

    def pack(self):
        if len(self.payload) < AFDX.MIN_PAYLOAD_LEN:
            raise ValueError("Minimum Payload of {} bytes".format(AFDX.MIN_PAYLOAD_LEN))

        afdx_header = struct.pack(
            ">IHHBBBBH",
            AFDX.DSTMAC_CONST,
            self.vlink,
            (AFDX.SRCMAC_CONST >> 8),
            0,
            self.networkID,
            self.equipmentID,
            (self.interfaceID << 5),
            self.type,
        )

        packet = afdx_header + self.payload + struct.pack(">B", self.sequencenum)

        return packet

    def __eq__(self, other):
        if not isinstance(other, AFDX):
            return False

        for attr in ["type", "networkID", "interfaceID", "equipmentID", "vlink", "sequencenum", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True


class ICMP(object):
    """Class to handle ICMP packets

    >>> i = ICMP()
    >>> i.type = ICMP.TYPE_REPLY
    >>> i.code = 0
    >>> i.payload = bytes(1)
    >>> mypacket = i.pack()


    """

    TYPE_REPLY = 0x0
    TYPE_UNREACHABLE = 0x1
    TYPE_REDIRECT = 0x5
    TYPE_REQUEST = 0x8

    def __init__(self):
        self.type: int = 0
        self.code: int = 0
        self.request_id: int = 0
        self.request_sequence: int = 0
        self.payload = bytes()

    def pack(self) -> bytes:
        """
        Pack an ICMP object into a buffer of bytes
        """
        for attr in ("type", "code", "request_id", "request_sequence"):
            if type(getattr(self, attr)) != int:
                raise ValueError("Attribute {} is not an integer".format(attr))

        _hdr_no_checksum = struct.pack(">BBHHH", self.type, self.code, 0, self.request_id, self.request_sequence)
        _icmp_checksum = ip_calc_checksum(_hdr_no_checksum + self.payload)
        _hdr = _hdr_no_checksum[:2] + struct.pack("H", _icmp_checksum) + _hdr_no_checksum[4:]
        return _hdr + self.payload

    def unpack(self, buffer: bytes):
        raise NotImplementedError("Not implemented")


MOD = 1 << 16


def ones_comp_add16(num1, num2):
    result = num1 + num2
    return result if result < MOD else (result + 1) % MOD


class IGMPv3(object):
    """
    Simplified IGMPv3 support to generate queries and join requests
    """

    TYPE_QUERY = 0x11
    TYPE_MEMBERSHIP_REPORT = 0x22
    TYPE_REC_CHG_TO_EXCL_MODE = 4
    TYPE_REC_MODE_IS_EXCLUDE = 2

    IP_ADDR_QUERY = "224.0.0.1"
    IP_ADDR_JOIN = "224.0.0.22"
    MAC_ADDR = {IP_ADDR_QUERY: 0x1005E000001, IP_ADDR_JOIN: 0x1005E000016}

    def __init__(self):
        pass

    @staticmethod
    def membership_query() -> bytes:
        """
        Return a membership query
        :return:
        """
        query_type = 0x11
        max_resp_time_2p4s = 0x18
        cksum = 0xECD3
        mc_addr = "0.0.0.0"
        sts = 0x2
        qqic = 0x20
        num_src = 0x0
        return (
            struct.pack(">BBH", query_type, max_resp_time_2p4s, cksum)
            + socket.inet_aton(mc_addr)
            + struct.pack(">BBH", sts, qqic, num_src)
        )

    @staticmethod
    def join_groups(groups: typing.List[str]) -> bytes:
        """
        Join the specified groups
        :param groups: List of IP addresses
        :return: bytes
        """
        if len(groups) == 1:
            mode = IGMPv3.TYPE_REC_CHG_TO_EXCL_MODE
        else:
            mode = IGMPv3.TYPE_REC_MODE_IS_EXCLUDE

        _nochecksum = struct.pack(">BBHHH", IGMPv3.TYPE_MEMBERSHIP_REPORT, 0, 0, 0, len(groups))
        for _g in groups:
            _nochecksum += struct.pack(">BBH", mode, 0, 0) + socket.inet_aton(_g)

        all_16b_wds = struct.unpack(">{}H".format(int(len(_nochecksum) / 2)), _nochecksum)
        checksum = ~reduce(ones_comp_add16, all_16b_wds) & 0xFFFF
        return _nochecksum[:2] + struct.pack(">H", checksum) + _nochecksum[4:]


class ARP(object):
    """Minimal ARP class,. Very limited

    Args:
        object (_type_): _description_

    Returns:
        _type_: _description_

    >>> a = ARP()
    >>> a.dstip = "192.168.28.2"
    >>> b = ARP()
    >>> b.unpack(a.pack())
    >>> a == b
    True
    """

    OPER_REQUEST = 1
    OPER_REPLY = 2

    def __init__(self):
        self.hardware_type: int = 1
        self.protocol_type: int = Ethernet.TYPE_IPv4
        self.hardware_length: int = Ethernet.ADDR_LENGTH
        self.protocol_length: int = IPv4.ADDR_LENGTH
        self.operation: int = ARP.OPER_REQUEST
        self.srcmac: int = 0x0
        self.dstmac: int = 0x0
        self.srcip: str = "0.0.0.0"
        self.dstip: str = "0.0.0.0"

    def pack(self) -> bytes:
        """Convert ARP object into bytes

        Returns:
            bytes: bytes representation of the ARP packet
        """
        _raw = struct.pack(
            ">HHBBH", self.hardware_type, self.protocol_type, self.hardware_length, self.protocol_length, self.operation
        )
        _raw += pack48(self.srcmac) + socket.inet_aton(self.srcip) + pack48(self.dstmac) + socket.inet_aton(self.dstip)
        return _raw

    def unpack(self, buffer: bytes) -> None:
        """Conmvert the buffer into an ARP object

        Args:
            buffer (bytes): Buffer of bytes
        """
        (self.hardware_type, self.protocol_type, self.hardware_length, self.protocol_length, self.operation) = (
            struct.unpack_from(">HHBBH", buffer)
        )
        self.srcmac = unpack48(buffer[8:14])
        self.srcip = socket.inet_ntoa(buffer[14:18])
        self.dstmac = unpack48(buffer[18:24])
        self.dstip = socket.inet_ntoa(buffer[24:28])

    def __repr__(self):
        return f"ARP: HW_type={self.hardware_type} ProtoType={self.protocol_type}, SRC={self.srcmac}/{self.srcip} DST={self.dstmac}/{self.dstip}"

    def __eq__(self, other: ARP) -> bool:
        if not isinstance(other, ARP):
            return False
        for attr in [
            "hardware_type",
            "protocol_type",
            "hardware_length",
            "protocol_length",
            "operation",
            "srcmac",
            "dstmac",
            "srcip",
            "dstip",
        ]:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True


def combine_ip_fragments(packets: typing.List[IP]) -> IP:
    """
    Combine the lists of fragmented IP packets into one IP packet

    """
    ident = None
    # Check first the we have the correct imput
    for packet in packets:
        if not isinstance(packet, IP):
            raise Exception("packet is not of type IP")
        if ident is not None:
            if packet.id != ident:
                raise Exception("All packets should have the same ID field")
        ident = packet.id

    # Create the IP packet we will return
    combined_ip = IP()
    combined_ip.flags = 0x0
    combined_ip.fragment_offset = 0x0
    combined_ip.payload = bytes()
    # Create the header from the first fragment. Order the packets via the fragment offsets
    # I don't verify the offsets but blindly combine them. TODO: Improve
    for idx, packet in enumerate(sorted(packets, key=lambda x: x.fragment_offset)):
        if idx == 0:
            for attr in ["srcip", "dstip", "protocol", "version", "ihl", "dscp", "id", "ttl"]:
                setattr(combined_ip, attr, getattr(packet, attr))
        # Add the payload
        combined_ip.payload += packet.payload

    # Return the IP packet
    return combined_ip



import struct
from collections import defaultdict
from enum import IntEnum


class _ComputerGeneratedData(object):
    """Base class for the computer generated classes"""

    def __init__(self):
        self._csdw = 0  # type: int
        self.payload = b""

    def pack(self):
        return struct.pack("<I", self._csdw) + self.payload

    def unpack(self, buffer):
        (self._csdw,) = struct.unpack_from("<I", buffer, 0)
        self.payload = buffer[4:]


class ComputerGeneratedFormat0(_ComputerGeneratedData):
    """Class to handled ComputerGeneratedFormat0 payloads"""

    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "Computer Generate Format 0. Data Length={}".format(len(self.payload))


FRMT_ASCII = 0
FRMT_XML = 1
FRMT = defaultdict(lambda: "undefined")
FRMT[FRMT_ASCII] = "ASCII Format"
FRMT[FRMT_XML] = "XML Format"

SRCC = defaultdict(lambda: "undefined")
SRCC_NO_CHANGE = 0
SRCC_CHANGE = 1
SRCC[SRCC_NO_CHANGE] = "No Change to Setup"
SRCC[SRCC_CHANGE] = "Change to Setup"


class RCCVER(IntEnum):
    IRIG_106_07 = 0x7
    IRIG_106_09 = 0x8
    IRIG_106_11 = 0x9
    IRIG_106_13 = 0xA
    IRIG_106_15 = 0xB
    IRIG_106_17 = 0xC
    IRIG_106_19 = 0xD
    IRIG_106_22 = 0xE

    @classmethod
    def _missing_(cls, value):
        return cls.IRIG_106_07


class ComputerGeneratedFormat1(_ComputerGeneratedData):
    """Class to handled ComputerGeneratedFormat1 payloads

    >>> c = ComputerGeneratedFormat1()
    >>> c.payload = bytes(10)
    >>> buffer = c.pack()

    """

    def __init__(self):
        super().__init__()
        self.frmt: int = FRMT_ASCII  # setup record format.
        self.srcc: int = SRCC_NO_CHANGE  # Setup Record Configuration Change (SRCC)
        self.rccver: int = RCCVER.IRIG_106_07  # RCC 106 Version (RCCVER

    def unpack(self, buffer: bytes) -> None:
        super().unpack(buffer)
        self.frmt = (self._csdw >> 9) & 0x1
        self.srcc = (self._csdw >> 8) & 0x1
        self.rccver = RCCVER(self._csdw & 0xFF)
        return True

    def pack(self) -> bytes:
        self._csdw = (self.frmt << 9) + (self.srcc << 8) + self.rccver
        return struct.pack("<I", self._csdw) + self.payload

    def __repr__(self):
        return "Computer Generate Format 1.FRMT={} SRCC={} RCCVER={} Data Length={}".format(
            FRMT[self.frmt], SRCC[self.srcc], self.rccver, len(self.payload)
        )



import unittest
import AcraNetwork.IRIG106.Chapter7 as ch7
import AcraNetwork.SimpleEthernet as eth
import AcraNetwork.Pcap as pcap
import os
import copy
import struct
import logging
from pstats import Stats
import cProfile
import typing
import random


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

logging.basicConfig(level=logging.INFO, format="%(levelname)s:%(funcName)s:%(lineno)s:%(message)s")


def buf_generator(count, llp_count=0):
    running_count = 0
    while running_count < count:
        buf_len = os.urandom((running_count + 1) * 128)
        running_count += 1
        if running_count <= llp_count:
            low_latency = True
        else:
            low_latency = False
        yield buf_len, low_latency


def ptfr_to_pcm_frame(
    count: int,
):
    remainder = bytes()
    pcm_frame_len = 1024
    offset_ptfr = 30
    zero_buf = struct.pack(">B", 0) * offset_ptfr
    pcm_frame = zero_buf
    for ptfr in ch7.datapkts_to_ptfr(buf_generator(count)):
        pcm_frame += ptfr.pack()
        if len(pcm_frame) >= pcm_frame_len:
            remainder = pcm_frame[pcm_frame_len:]
            pcm_frame = pcm_frame[:pcm_frame_len]
            yield pcm_frame
            pcm_frame = zero_buf + remainder


class TestCaseCh7(unittest.TestCase):
    def test_basic(self):
        ch7_pkt = ch7.PTFR()
        ch7_pkt.length = 132
        ch7_pkt.llp = 0
        remainder = bytes()
        while remainder == bytes():
            ch7_pd = ch7.PTDP()
            ch7_pd.content = ch7.PTDPContent.ETHERNET_MAC
            ch7_pd.fragment = ch7.PTDPFragment.COMPLETE
            ch7_pd.payload = os.urandom(58)
            remainder = ch7_pkt.add_payload(ch7_pd.pack())
            # print("Packet added")
        buf = ch7_pkt.pack()
        # print("Len Buf={} Len Rem={}".format(len(buf), len(remainder)))
        # Check the comparsion
        ch7_pkt_copy = copy.deepcopy(ch7_pkt)
        self.assertTrue(ch7_pkt == ch7_pkt_copy)

        ch7_unpack = ch7.PTFR()
        ch7_unpack.length = 132
        self.assertTrue(ch7_unpack.unpack(buf))
        self.assertTrue(ch7_unpack == ch7_pkt)

    def test_partial_fill(self):
        ch7_pkt = ch7.PTFR()
        ch7_pkt.length = 140
        ch7_pkt.llp = 0
        remainder = bytes()
        while remainder == bytes():
            ch7_pd = ch7.PTDP()
            ch7_pd.content = ch7.PTDPContent.ETHERNET_MAC
            ch7_pd.fragment = ch7.PTDPFragment.COMPLETE
            ch7_pd.payload = os.urandom(58)
            remainder = ch7_pkt.add_payload(ch7_pd.pack())
            # print("Packet added")
        buf = ch7_pkt.pack()
        ch7_unpack = ch7.PTFR()
        ch7_unpack.length = 140
        self.assertTrue(ch7_unpack.unpack(buf))
        # print(repr(ch7_pkt))
        # print(repr(ch7_unpack))
        self.assertEqual(52, len(remainder))
        self.assertTrue(ch7_unpack == ch7_pkt)

    def test_comparsion(self):
        ch7_pd = ch7.PTDP()
        ch7_pd.content = ch7.PTDPContent.ETHERNET_MAC
        ch7_pd.fragment = ch7.PTDPFragment.COMPLETE
        ch7_pd.payload = os.urandom(60)
        ch_pd_copy = copy.deepcopy(ch7_pd)
        self.assertTrue(ch7_pd == ch_pd_copy)


class TestGenerators(unittest.TestCase):
    def test_ptdp_generator(self):
        for ptdp_pkt in ch7.datapkts_to_ptdp(buf_generator(5)):
            # print(repr(ptdp_pkt))
            self.assertIsInstance(ptdp_pkt, ch7.PTDP)

    def test_ptfr_generator(self):
        remainder = bytes()
        for ptfr in ch7.datapkts_to_ptfr(buf_generator(5), ptfr_len=200):
            # print(repr(ptfr))
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 800
            ch7_pkt.unpack(ptfr.pack())
            for p, remainder, e in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    # print(repr(p))
                    self.assertEqual(0, p.length % 128)


class TestRealEthernet(unittest.TestCase):
    def setUp(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def tearDown(self):
        p = Stats(self.pr)
        p.sort_stats("cumtime")
        # p.print_stats()

    pkts_sent = []

    @staticmethod
    def eth_gen(count, low_latency_pkts=None, size_mult=128):
        """

        :param count:
        :param low_latency_pkts:
        :param size_mult:
        :rtype:  collections.Iterable[bytes, bool]
        """

        pf = pcap.Pcap(THIS_DIR + "/generated_eth.pcap", mode="w")
        r = pcap.PcapRecord()
        for i in range(1, count + 1):
            e = eth.Ethernet()
            e.dstmac = 0xABABABABABAB
            e.srcmac = 0xDCDCDCDCDCDC
            e.type = e.TYPE_IP
            ip = eth.IP()
            ip.srcip = "192.168.28.3"
            ip.dstip = "192.168.28.1"
            u = eth.UDP()
            u.dstport = i
            u.srcport = i
            if size_mult == 0:
                u.payload = struct.pack(">{}H".format(64), *([0xA5A5] * (64)))
            else:
                u.payload = struct.pack(">{}H".format(i * size_mult), *([0xA5A5] * (i * size_mult)))
            ip.payload = u.pack()
            e.payload = ip.pack()

            if low_latency_pkts is None:
                low_latency = False
            elif i in low_latency_pkts:
                low_latency = True
            else:
                low_latency = False
            r.setCurrentTime()
            r.payload = e.pack()
            pf.write(r)
            TestRealEthernet.pkts_sent.append(e.pack())
            yield e.pack(), low_latency
        pf.close()

    def test_eth_in_packets(self):
        pf = pcap.Pcap(THIS_DIR + "/gen.pcap", mode="w")
        r = pcap.PcapRecord()
        remainder = bytes()
        eth_p = bytes()
        pkt_count = 1
        for ptfr in ch7.datapkts_to_ptfr(TestRealEthernet.eth_gen(3), ptfr_len=200):
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 200
            ch7_pkt.unpack(ptfr.pack())
            for p, remainder, e in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    eth_p += p.payload
                    if p.fragment == ch7.PTDPFragment.COMPLETE or p.fragment == ch7.PTDPFragment.LAST:
                        r.setCurrentTime()
                        r.payload = eth_p
                        # Verify the size of the packets and it's probably good. I could unpack them here too
                        self.assertEqual(
                            pkt_count * 2 * 128
                            + eth.UDP.UDP_HEADER_SIZE
                            + eth.IP.IP_HEADER_SIZE
                            + eth.Ethernet.HEADERLEN,
                            len(eth_p),
                        )
                        pkt_count += 1
                        pf.write(r)
                        eth_p = bytes()

                    # self.assertEqual(0, p.length % 128)
                    # self.assertTrue(p.length <= 1024)
        pf.close()

    def test_eth_in_packets_low_latency(self):
        # logging.basicConfig(level=logging.DEBUG)
        pf = pcap.Pcap(THIS_DIR + "/captured_llp.pcap", mode="w")
        r = pcap.PcapRecord()
        remainder = bytes()
        eth_p = bytes()
        pkt_count = 1
        ptfr_idx = 0
        ptdp_idx = 0
        pkt_size_mult = 16
        for ptfr in ch7.datapkts_to_ptfr(
            TestRealEthernet.eth_gen(20, low_latency_pkts=[2, 4, 10], size_mult=8), ptfr_len=400
        ):
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 400
            ch7_pkt.unpack(ptfr.pack())
            # b = open("gen_llp_{}.bin".format(ptfr_idx), mode="wb")
            ptfr_idx += 1
            # b.write(ptfr.pack())
            # b.close()
            # print(repr(ptfr))
            for p, remainder, e in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    # print(repr(p))
                    eth_p += p.payload
                    if p.fragment == ch7.PTDPFragment.COMPLETE or p.fragment == ch7.PTDPFragment.LAST:
                        if ptdp_idx == 0:
                            self.assertEqual(p.low_latency, True)
                        r.setCurrentTime()
                        r.payload = eth_p
                        # Verify the size of the packets and it's probably good. I could unpack them here too
                        # self.assertEqual(2*64+eth.UDP.UDP_HEADER_SIZE+eth.IP.IP_HEADER_SIZE+eth.Ethernet.HEADERLEN, len(eth_p))
                        if p.content != ch7.PTDPContent.FILL:
                            if eth_p not in TestRealEthernet.pkts_sent:
                                self.assertTrue(False)
                            pkt_count += 1
                        pf.write(r)
                        eth_p = bytes()
                    ptdp_idx += 1
        self.assertEqual(pkt_count, 11)

        # self.assertEqual(0, p.length % 128)
        # self.assertTrue(p.length <= 1024)
        pf.close()

    def test_llc(self):
        llc_count = 0
        fill_count = 0
        mac_count = 0
        remainder = b""
        for i in range(3, 6):
            f = open(THIS_DIR + "/ptfr_{}.bin".format(i), "rb")
            ptfr_data = f.read()
            f.close()

            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = len(ptfr_data)
            eth_p = b""
            ch7_pkt.unpack(ptfr_data)
            # print repr(ch7_pkt)

            for p, remainder, e in ch7_pkt.get_aligned_payload(remainder):
                if p is not None:
                    if p.content != ch7.PTDPContent.FILL:
                        mac_count += 1
                        # print(repr(p))
                    if p.low_latency:
                        llc_count += 1
                    if p.content == ch7.PTDPContent.FILL:
                        fill_count += 1
        # print("{} {} {}".format(llc_count, fill_count, mac_count))
        self.assertEqual(4, llc_count)
        self.assertEqual(164, fill_count)
        self.assertEqual(4, mac_count)


def get_pkts(some_low_latency: bool = False, max_len: int = 178) -> typing.Generator[tuple[bytes, bool], None, None]:

    count = 0
    while True:
        # pkt_len = random.randint(2, 180)
        pkt_len = (count % max_len) + 2
        paylaod_int = [pkt_len] + [count] * (pkt_len - 1)
        payload = struct.pack(f">{pkt_len}Q", *paylaod_int)
        count += 1
        llc_pkts = [True, False, False, False, False, False, False]
        # llc_pkts = [True, True, True, True, True, True, True]
        if some_low_latency:
            low_latency = llc_pkts[count % 7]
        else:
            low_latency = False
        logging.debug(f"TX: Generated payload of length {pkt_len*8} count={count}")
        yield payload, low_latency


def get_pcm_frame(offset_ptfr: int = 0, some_low_latency: bool = False, max_len: int = 178):
    pcm_frame_len = 1024
    ptfr_len = pcm_frame_len - offset_ptfr - 4
    zero_buf = struct.pack(">B", 0) * offset_ptfr
    for ptfr in ch7.datapkts_to_ptfr(get_pkts(some_low_latency, max_len), ptfr_len=ptfr_len):
        pcm_frame = zero_buf + ptfr.pack()
        logging.debug(f"TX pcm_frame_len={len(pcm_frame)} ptfr_len={ptfr_len}")
        yield pcm_frame


def missing_elements(L):
    start, end = L[0], L[-1]
    return sorted(set(range(start, end + 1)).difference(L))


class TestRandomSizedDecom(unittest.TestCase):
    def test_no_llc(self):
        offset = 0
        first_PTFR = True
        eth_p = bytes()
        prev_eth_count = None
        remainder = None
        count = 0
        for frame in get_pcm_frame(offset, some_low_latency=False):
            ch7_pkt = ch7.PTFR()
            ch7_buffer = frame[offset:]
            ch7_pkt.length = len(ch7_buffer)
            ch7_pkt.unpack(ch7_buffer)
            count += 1
            if count > 10000:
                return

            for p, remainder, e in ch7_pkt.get_aligned_payload(first_PTFR, remainder):
                first_PTFR = False
                if p is not None:
                    if p.length != 0:
                        if p.fragment == ch7.PTDPFragment.COMPLETE or p.fragment == ch7.PTDPFragment.LAST:
                            eth_p += p.payload
                            logging.debug(repr(p))
                            self.assertGreaterEqual(len(eth_p), 16)
                            (expected_len, count) = struct.unpack_from(">QQ", eth_p, 0x0)
                            logging.debug(f"RX payload count={count} len={len(eth_p)}")
                            if prev_eth_count is not None:
                                if prev_eth_count + 1 != count:
                                    self.assertEqual(prev_eth_count + 1, count)
                                self.assertEqual(expected_len * 8, len(eth_p))

                            prev_eth_count = count
                            eth_p = bytes()

    def test_some_llc(self):
        offset = 0
        first_PTFR = True
        eth_p = bytes()
        remainder = None
        golay = ch7.Golay.Golay()
        count = 0
        numbers_found = []
        for frame in get_pcm_frame(offset, some_low_latency=True, max_len=50):
            ch7_pkt = ch7.PTFR(golay)
            ch7_buffer = frame[offset:]
            ch7_pkt.length = len(ch7_buffer)
            ch7_pkt.unpack(ch7_buffer)
            count += 1
            if count > 10000:
                return

            for p, remainder, e in ch7_pkt.get_aligned_payload(first_PTFR, remainder):
                first_PTFR = False
                if p is not None:
                    if p.length != 0:
                        if p.fragment == ch7.PTDPFragment.COMPLETE or p.fragment == ch7.PTDPFragment.LAST:
                            eth_p += p.payload
                            logging.debug(repr(p))
                            if p.content != ch7.PTDPContent.FILL:
                                self.assertGreaterEqual(len(eth_p), 16)
                                (expected_len, count) = struct.unpack_from(">QQ", eth_p, 0x0)
                                logging.debug(f"RX payload count={count} len={len(eth_p)}")
                                numbers_found.append(count)
                                self.assertEqual(expected_len * 8, len(eth_p))

                            eth_p = bytes()

        numbers_found.sort()
        self.assertEqual(missing_elements(numbers_found), [])


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import struct
import typing


class ARINC429DataWord(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles ARINC-429 packets
    https://www.irig106.org/docs/106-22/chapter11.pdf    11.2.8.1

    This object is generally encapsulated in :class:`ARINC429DataPacket` objects

    """

    LO_SPEED = 0  #: Bus speed constant
    HI_SPEED = 1  #: Bus speed constant

    HDR_FORMAT = "<I"

    def __init__(self):
        self.gaptime: int = 0  #: The gap time from the beginning of the preceding bus word (regardless of bus) to the
        # beginning of the current bus word in 0.1-us increments
        self.format_error: bool = False  #: Format error has occurred
        self.parity_error: bool = False  #: Parity error has occurred
        self.bus_speed: int = ARINC429DataWord.LO_SPEED  #: Arinc bus speed
        self.bus: int = 0  #: Bus number index from 0
        self.payload: bytes = bytes()  #: ARINC word as a string payload

    def pack(self) -> bytes:
        """
        Pack the ARINC-429 data packet object into a binary buffer

        """
        _ipdh = (
            (self.bus << 24)
            + (self.format_error << 23)
            + (self.parity_error << 22)
            + (self.bus_speed << 21)
            + (self.gaptime)
        )
        hdr = struct.pack(ARINC429DataWord.HDR_FORMAT, _ipdh)

        return hdr + self.payload

    def unpack(self, buffer: bytes):
        """
        Unpack a string buffer into an ARINC-429 data packet object

        :param buffer: A string buffer representing an ARINC-429 data  packet
        :type buffer: bytes
        :rtype: None
        """
        (_ipdh,) = struct.unpack_from(ARINC429DataWord.HDR_FORMAT, buffer)
        self.payload = buffer[struct.calcsize(ARINC429DataWord.HDR_FORMAT) :]
        self.bus = _ipdh >> 24
        self.format_error = bool((_ipdh >> 23) & 0x1)
        self.parity_error = bool((_ipdh >> 22) & 0x1)
        self.bus_speed = (_ipdh >> 21) & 0x1
        self.gaptime = _ipdh & 0xFFFFF

        return True

    def __eq__(self, other):
        if not isinstance(other, ARINC429DataWord):
            return False

        _match_att = ("gaptime", "format_error", "parity_error", "bus_speed", "bus", "payload")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "ARINCData: GapTime={} FormatError={} ParityError={} BusSpeed={} Bus={}".format(
            self.gaptime, self.format_error, self.parity_error, self.bus_speed, self.bus
        )


class ARINC429DataPacket(object):
    """
    Data Packet Format. Contains a list of Arinc Data Words

    :type msgcount: int
    :type arincwords: list[ARINC429DataWord]


    >>> from base64 import b64decode
    >>> from AcraNetwork.IRIG106.Chapter11 import Chapter11
    >>> data = b64decode('JesACogAAABkAAAABgDEOFcwAAAAADJfCOISAAAAAAAAAPwADAAAAAAAAAC+AgBgEA4AAMEFAGAQDgAAbQsAQBAOAAB5DwAAEA4AAEITAMAQDgAAJBQAQBAOAAD3GQCAEA4AAAQdAMAQDgAAZSMAABAOAADFJABAEA4AAJ8pAGAQDgAAgSwAwA==')
    >>> c = Chapter11()
    >>> c.unpack(data)
    True
    >>> arinc_p = ARINC429DataPacket()
    >>> arinc_p.unpack(c.payload)
    True
    >>> print(arinc_p)
    ARINCPayload: MessageCount=12
      ARINCData: GapTime=0 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
      ARINCData: GapTime=4110 FormatError=False ParityError=False BusSpeed=0 Bus=0
    <BLANKLINE>
    >>> c = Chapter11()
    >>> arinc_p = ARINC429DataPacket()
    >>> arinc_dw = ARINC429DataWord()
    >>> arinc_dw.payload = bytes(1)
    >>> arinc_p.append(arinc_dw)
    >>> c.payload = arinc_p.pack()

    """

    def __init__(self):
        self.msgcount: int = 0  #: The number ofARINC-429 words included in the packet.
        self.arincwords: typing.List[ARINC429DataWord] = []  #: List of :class:`ARINC429DataWord`

    def pack(self) -> bytes:
        """
        Pack the ARINC-429 data packet object into a binary buffer

        :rtype: bytes
        """
        ret_str = struct.pack("<HH", self.msgcount, 0)
        for a in self.arincwords:
            ret_str += a.pack()

        return ret_str

    def append(self, dataword: ARINC429DataWord):
        """Add the arinc dataword to thie packet

        Args:
            dataword (ARINC429DataWord): The word to add
        """
        self.arincwords.append(dataword)

    def unpack(self, buffer: bytes) -> bool:
        """
        Unpack a string buffer into an ARINC-429 data packet object

        :param buffer: A string buffer representing an ARINC-429 data  packet
        :type buffer: bytes
        :rtype: None
        """
        CH_SPECIFIC_HDR_LEN = 4
        ARINC_WORD_LEN = 8
        (self.msgcount, _res) = struct.unpack_from("<HH", buffer)
        exp_msg = (len(buffer) - CH_SPECIFIC_HDR_LEN) // ARINC_WORD_LEN
        for msg_idx in range(exp_msg):
            offset = (msg_idx * ARINC_WORD_LEN) + CH_SPECIFIC_HDR_LEN
            arinc_data = ARINC429DataWord()
            arinc_msg_word_buffer = buffer[offset : offset + ARINC_WORD_LEN]
            arinc_data.unpack(arinc_msg_word_buffer)
            self.arincwords.append(arinc_data)

        if self.msgcount != len(self.arincwords):
            raise Exception(
                "The ARINC Message Count={} does not match number of messages in the packet={}".format(
                    self.msgcount, len(self.arincwords)
                )
            )
        return True

    def __eq__(self, other: ARINC429DataPacket):
        if not isinstance(other, ARINC429DataPacket):
            return False

        _match_att = ("msgcount", "arincwords")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str = "ARINCPayload: MessageCount={}\n".format(self.msgcount)

        for a in self.arincwords:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self) -> ARINC429DataWord:
        if self._index < len(self.arincwords):
            _dw = self.arincwords[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.arincwords)

    def __getitem__(self, key):
        return self.arincwords[key]



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
"""
ADAU generates pcap files which contain a chapter10 UDP header, and timestamps as PTP secondard header.
The application creates a new ch10 file, adds in the TMATs file, adds in Time packets ever second
and then converts all PTP timestamps to RTC
"""
__author__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Prototype"

import AcraNetwork.IRIG106.Chapter11 as ch10
import AcraNetwork.Chapter10.Chapter10UDP as ch10udp
import AcraNetwork.IRIG106.Chapter11.UART as ch10uart
import AcraNetwork.IRIG106.Chapter11.PCM as ch10pcm
import AcraNetwork.IRIG106.Chapter11.MILSTD1553 as ch10mil
import AcraNetwork.IRIG106.Chapter11.TimeDataFormat as ch10time
import AcraNetwork.IRIG106.Chapter11.ComputerData as chcomputer
import AcraNetwork.IRIG106.Chapter11.ARINC429 as charinc
from AcraNetwork.IRIG106.Chapter10 import FileParser
from AcraNetwork import endianness_swap
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as eth
from AcraNetwork import __version__
import argparse
import sys
from dataclasses import dataclass
import logging
from collections import defaultdict
import struct
from AcraNetwork.Chapter10 import (
    TS_CH4,
    TS_IEEE1558,
    PTPTime,
    RTCTime,
    DataType,
)
import typing
import glob
import re


# Handle integers and hex representations for command line argu,ents
def auto_int(x):
    return int(x, 0)


def natural_sort(mylist):
    """
    Natural sort of a list.
    :param l:
    :return:
    """
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(mylist, key=alphanum_key)


@dataclass
class PktDetails:
    """Details to track for each packet"""

    channelid: int
    datatype: DataType

    def __hash__(self) -> int:
        return hash((self.channelid, self.datatype))


def create_parser():
    # ----------------------------------
    # Setup the command line parser
    # ----------------------------------
    parser = argparse.ArgumentParser(
        description="Convert an ADAU pcap to chapter10 file. User must provide the TMATs file. The packet format is converted to RTC time"
    )
    parser.add_argument("--pcap", required=False, help="Select a single input file")
    parser.add_argument("--pcapfolder", required=False, help="Select a folder and parse all pcap files in the folder")
    parser.add_argument(
        "--pcmsyncword",
        type=auto_int,
        required=False,
        default=None,
        help="The sync word for all PCM payloads found. Set to 0xFE6B_2840 for the standard Sync word",
    )
    parser.add_argument(
        "--timeid",
        type=int,
        required=False,
        default=512,
        help="The channel ID of the time packet",
    )
    parser.add_argument("--ch10", required=True, help="The out chapter 10 file")
    parser.add_argument(
        "--pcapdebug", required=False, default=None, help="Create a debug pcap file which matches the chapter 10 file"
    )
    parser.add_argument("--tmats", required=True, help="The TMATS input file")
    parser.add_argument("--verbose", action="store_true", required=False, default=False, help="verbose mode")
    parser.add_argument("--debug", action="store_true", required=False, default=False, help="debug mode")
    parser.add_argument("--version", action="version", version=f"{__version__}")
    return parser


def encapsulate_ch10_ptk(ch10payload: bytes) -> bytes:
    """
    Encapsulte the udp payload in an Ethernet packet for debug puprposes
    """
    ethpkt = eth.Ethernet()
    ethpkt.dstmac = 0x01005E000001
    ethpkt.srcmac = 0x000C4DAC7AAA
    ethpkt.type = eth.Ethernet.TYPE_IP
    #
    ippkt = eth.IP()
    ippkt.dstip = "235.0.0.2"
    ippkt.srcip = "127.0.0.1"
    # Stick a UDP packet in the payload
    udppkt = eth.UDP()
    udppkt.dstport = 51001
    udppkt.srcport = 51001

    fixed_ch10udp = struct.pack(">HH", 0x17C, 0x1911)
    udppkt.payload = fixed_ch10udp + ch10payload
    # packet the udp packet into the ethernet payload
    ippkt.payload = udppkt.pack()
    ethpkt.payload = ippkt.pack()
    return ethpkt.pack()


def encapsulate_tmats(tmats_file: str, rtctime: int) -> ch10.Chapter11:
    """Wrap the tmats data in a ch10 file for writing to a ch10 file

    Args:
        tmats_data (bytes): _description_

    Returns:
        ch10.Chapter10: _description_
    """
    with open(tmats_file, mode="rb") as f:
        tmats = f.read()
        c = ch10.Chapter11()
        c.channelID = 0
        c.sequence = 0
        c.packetflag = 0
        c.datatype = DataType.COMPUTER_FORMAT_1
        c.relativetimecounter = rtctime
        ctmats = chcomputer.ComputerGeneratedFormat1()
        ctmats.payload = tmats
        c.payload = ctmats.pack()
        return c


def get_ch10_time(ptptime: PTPTime, sequence: int = 0, channelid: int = 512) -> ch10.Chapter11:
    """Return a time packet that maps the seconds / nanoseconds to the rtctime

    Args:
        seconds (int): _description_
        nanoseconds (int): _description_
        rtctime (int): _description_

    Returns:
        ch10.Chapter10: _description_
    """
    c = ch10.Chapter11()
    c.channelID = channelid
    c.sequence = sequence
    c.packetflag = 0
    c.datatype = DataType.TIMEFORMAT_1
    c.relativetimecounter = ptptime.to_pinksheet_rtc()
    time_pkt = ch10time.TimeDataFormat1()
    time_pkt.ptptime = ptptime
    c.payload = time_pkt.pack()
    return c


def clone_ch10_payload(
    original_buffer: bytes,
    datatype: int,
    pcmsyncword: typing.Optional[int] = None,
    minor_frame_size_bytes: typing.Optional[int] = None,
) -> typing.Tuple[bytes, typing.Optional[int]]:
    """Return the payload with the secondary header removed

    Args:
        original_buffer (bytes): _description_

    Returns:
        tuple of bytes and minor frame length
    """
    if datatype == DataType.ARINC429:
        p = charinc.ARINC429DataPacket()
        p.unpack(original_buffer)
        for arincpay in p:
            arincpay.payload = endianness_swap(arincpay.payload, 4)
        return p.pack(), None

    elif datatype == DataType.MILSTD1553:
        p = ch10mil.MILSTD1553DataPacket(TS_IEEE1558)
        p.unpack(original_buffer)
        # Go through each message and conver the timestamp to RTC
        for milpayload in p:
            milpayload.ipts = RTCTime(milpayload.ipts.to_pinksheet_rtc())
            # Swap the endianness of the message
            milpayload.message = endianness_swap(milpayload.message)
        return p.pack(), None

    elif datatype == DataType.UART:
        p = ch10uart.UARTDataPacket(TS_IEEE1558)
        p.unpack(original_buffer)
        # Go through each message and conver the timestamp to RTC
        for dataword in p:
            if dataword.ipts is not None:
                dataword.ipts = RTCTime(dataword.ipts.to_pinksheet_rtc())
        return p.pack(), None

    elif datatype == DataType.PCM:
        p = ch10pcm.PCMDataPacket(
            ipts_source=TS_IEEE1558, syncword=pcmsyncword, minor_frame_size_bytes=minor_frame_size_bytes
        )
        p.unpack(original_buffer)
        # p.channel_specific_word = 0x0
        for frame in p:
            # logging.debug(f"PCMFrame={frame}")
            if frame.ipts is not None:
                frame.ipts = RTCTime(frame.ipts.to_pinksheet_rtc())
            # Swap the endianness of the message
            frame.minor_frame_data = endianness_swap(frame.minor_frame_data)
        return p.pack(), p.minor_frame_size_bytes

    elif datatype == DataType.ANALOG:
        return original_buffer, None

    else:
        raise Exception(f"Data type {datatype} not supported")


def clone_ch10(
    original_ch10: ch10.Chapter11,
    syncword: typing.Optional[int] = None,
    minor_frame_size_bytes: typing.Dict[int, typing.Optional[int]] = {},
) -> ch10.Chapter11:
    """Clone the ch10 packet but remove the secondary header

    Args:
        original_ch10 (ch10.Chapter10): _description_

    Returns:
        ch10.Chapter10: _description_
    """
    new_ch10 = ch10.Chapter11()
    new_ch10.syncpattern = original_ch10.syncpattern
    new_ch10.channelID = original_ch10.channelID
    new_ch10.datatypeversion = original_ch10.datatypeversion
    new_ch10.sequence = original_ch10.sequence
    new_ch10.packetflag = original_ch10.packetflag
    new_ch10.datatype = original_ch10.datatype

    # logging.debug(f"Fkags={original_ch10.packetflag:#0X}")
    if original_ch10.packetflag & ch10.Chapter11.PKT_FLAG_SEC_HDR_TIME != 0:
        new_ch10.packetflag &= 0x33
        logging.debug(f"Converting timestamps and overwrite packet flag {new_ch10.packetflag:#0X}")
        new_ch10.relativetimecounter = original_ch10.ptptime.to_pinksheet_rtc()
    else:
        new_ch10.relativetimecounter = original_ch10.relativetimecounter
    new_ch10.ts_source = ch10.TS_RTC
    new_ch10.payload, framesize = clone_ch10_payload(
        original_ch10.payload, new_ch10.datatype, syncword, minor_frame_size_bytes[new_ch10.channelID]
    )
    minor_frame_size_bytes[new_ch10.channelID] = framesize
    return new_ch10


def update_stats(pkt_details: typing.Dict[PktDetails, int], detail: PktDetails) -> None:
    """Keep a count of the packets received

    Args:
        pkt_details (dict[PktDetails, int]): _description_
        detail (PktDetails): _description_
    """
    if detail not in pkt_details:
        pkt_details[detail] = 1
        print(f"Found data type {detail.datatype.name} on channelID {detail.channelid}")
    else:
        pkt_details[detail] += 1


def print_stats(pkt_details: typing.Dict[PktDetails, int]):
    """Print the stats at the end

    Args:
        pkt_details (dict[PktDetails, int]): _description_
    """
    for detail, count in pkt_details.items():
        print(f"Packet Type={detail.datatype.name} channelID={detail.channelid} count={count}")


def main(args):
    if args.pcap:
        pcap_files = [args.pcap]
    elif args.pcapfolder:
        pcap_files = natural_sort(glob.glob("{}/*.pcap".format(args.pcapfolder)))
    else:
        print("ERROR: Please specify an input pcap file or an input pcap folder")
        return 1

    if args.pcapdebug is not None:
        pf_tmp = pcap.Pcap(args.pcapdebug, mode="w")
    else:
        pf_tmp = None
    fp = FileParser.FileParser(args.ch10, mode="wb")

    CH10_DATA_OFFSET = 0x2A
    CH10_DATA_LEN_MIN = 30

    prev_time = None
    time_sequnece = 0
    pkt_type_count = {}
    aligned_to_10_ms = False
    # The key is the channelID. So keep track of the minor frames per channelID
    minor_frame_size_bytes: typing.Dict[int, int | None] = defaultdict(lambda: None)

    with fp as ch10file:
        # Generate the TMATs ch10 packet and write to the output file
        new_ch10_pkt = None
        for pfile in pcap_files:
            pf = pcap.Pcap(pfile, mode="r")
            print(f"Converting {pfile}")
            for idx, record in enumerate(pf):
                # Pull out the ethernet packet
                eth_pkt = record.payload

                # Debug output record
                newrec = pcap.PcapRecord()
                newrec.sec = record.sec
                newrec.usec = record.usec

                # Sanity check on the input packet
                if len(eth_pkt) > CH10_DATA_OFFSET + CH10_DATA_LEN_MIN:
                    logging.debug(f"Reading record {idx}")
                    # Get the UDP payload (ie the chapter10 packet)
                    pkt_payload = eth_pkt[CH10_DATA_OFFSET:]
                    ch10udp_pkt = ch10udp.Chapter10UDP()
                    ch10_pkt = ch10.Chapter11()
                    # Unpack the ch10 UDP and ch10 packets
                    try:
                        ch10udp_pkt.unpack(pkt_payload)
                        ch10_pkt.unpack(ch10udp_pkt.payload)
                    except Exception as e:
                        logging.debug(f"Failed to unpacket record #{idx}. len_buf={len(pkt_payload)} Err={e}")
                        continue
                    else:
                        if prev_time is None:
                            time_in_ms = int(ch10_pkt.ptptime.nanoseconds // 1e6)
                            if time_in_ms % 10 == 0 or args.debug:
                                # Get the first timepacket and write it to the ch10 file
                                time_pkt = get_ch10_time(ch10_pkt.ptptime, time_sequnece, args.timeid)
                                # Now that we have the time, write the TMATs file
                                tmats_ch10 = encapsulate_tmats(args.tmats, time_pkt.relativetimecounter)
                                ch10file.write(tmats_ch10)
                                # Followed by the time packet
                                ch10file.write(time_pkt)

                                prev_time = ch10_pkt.ptptime
                                time_sequnece = (time_sequnece + 1) % 256
                                # write out the packets to the debug pcap
                                if pf_tmp is not None:
                                    newrec.payload = encapsulate_ch10_ptk(tmats_ch10.pack())
                                    pf_tmp.write(newrec)
                                    newrec.payload = encapsulate_ch10_ptk(time_pkt.pack())
                                    pf_tmp.write(newrec)
                                aligned_to_10_ms = True
                        elif prev_time is not None and ch10_pkt.ptptime is not None:
                            # Check if 1 second has elapsed since the previous time packet
                            dlt = ch10_pkt.ptptime - prev_time
                            if dlt >= PTPTime(1, 0):
                                time_pkt = get_ch10_time(ch10_pkt.ptptime, time_sequnece, args.timeid)
                                ch10file.write(time_pkt)
                                prev_time = ch10_pkt.ptptime  #
                                time_sequnece = (time_sequnece + 1) % 256
                                if pf_tmp:  # debug
                                    newrec.payload = encapsulate_ch10_ptk(time_pkt.pack())
                                    pf_tmp.write(newrec)
                        if aligned_to_10_ms:
                            # Convert the chapter10 packet to the RTC compatiable one
                            new_ch10_pkt = clone_ch10(ch10_pkt, args.pcmsyncword, minor_frame_size_bytes)

                            # Update the stats
                            update_stats(
                                pkt_type_count, PktDetails(new_ch10_pkt.channelID, DataType(new_ch10_pkt.datatype))
                            )
                            ch10file.write(new_ch10_pkt)

                        if pf_tmp is not None and aligned_to_10_ms and new_ch10_pkt is not None:  # debug
                            newrec.payload = encapsulate_ch10_ptk(new_ch10_pkt.pack())
                            pf_tmp.write(newrec)

    pf.close()
    print_stats(pkt_type_count)
    if pf_tmp is not None:
        pf_tmp.close()
    fp.close()


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
        pcmlogger = logging.getLogger("AcraNetwork.Chapter10.PCM")
        pcmlogger.setLevel(logging.DEBUG)

    ret = main(args)
    sys.exit(ret)



"""
.. module:: MPEG.PMT
    :platform: Unix, Windows
    :synopsis: Manage PMT packet formats https://en.wikipedia.org/wiki/Program-specific_information#

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import typing
import struct
from AcraNetwork.MPEGTS import MPEGPacket
from zlib import crc32
import logging

logger = logging.getLogger(__name__)


def bytes_to_ascii(buffer: bytes) -> str:
    r = ""
    for _b in buffer:
        r += chr(_b)
    return r


class DescriptorTag(object):
    """Descriptor tag fields in the PAT / PMT packets
    https://en.wikipedia.org/wiki/Program-specific_information#Descriptor

    Args:
        object (_type_): _description_

    Returns:
        _type_: _description_
    """

    FMT = ">BB"

    def __init__(self) -> None:
        self.tag: int = None
        self.data: bytes = bytes()

    def unpack(self, buffer: bytes) -> bytes:
        """Unpack the buffer into a descriptor and return any remainder

        Args:
            buffer (bytes): _description_

        Returns:
            : _description_
        """
        hdr_len = struct.calcsize(DescriptorTag.FMT)
        (self.tag, _len) = struct.unpack_from(DescriptorTag.FMT, buffer)
        self.data = buffer[hdr_len : hdr_len + _len]
        return buffer[hdr_len + _len :]

    def pack(self) -> bytes:
        """Convert the DescriptorTag into bytes"""
        return struct.pack(DescriptorTag.FMT, self.tag, len(self.data)) + self.data

    def __len__(self) -> int:
        if self.tag is None:
            return 0
        return 2 + len(self.data)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, DescriptorTag):
            return False
        if self.tag != value.tag or self.data != value.data:
            return False
        return True

    def __repr__(self) -> str:
        return f"Tag={self.tag:#0X} Length={len(self.data)}"


MIN_STREAM_LEN = 5


class PMTStream(object):
    """Elementary Stream specific data
    https://en.wikipedia.org/wiki/Program-specific_information#PMT_(Program_map_specific_data)

    Args:
        object (_type_): _description_

    Returns:
        _type_: _description_
    """

    FMT = ">BHH"

    def __init__(self) -> None:
        self.streamtype: int = 0x0
        self.elementary_pid: int = 0
        self.elementary_stream_descriptors: bytes = bytes()

    def unpack(self, buffer: bytes) -> bytes:
        """Extract a stream from the specified buffer and return the remainder bytes
        Iterate through a buffer until all the payload has been exhausted"""
        (self.streamtype, _pid, _len) = struct.unpack_from(PMTStream.FMT, buffer)
        self.elementary_pid = _pid & 0x1FFF
        es_len = _len & 0xFFF
        self.elementary_stream_descriptors = buffer[
            struct.calcsize(PMTStream.FMT) : struct.calcsize(PMTStream.FMT) + es_len
        ]
        # Return the remainder
        return buffer[es_len + struct.calcsize(PMTStream.FMT) :]

    def pack(self) -> bytes:
        """Convert the PMTSream into bytes

        Returns:
            bytes: _description_
        """

        hdr = struct.pack(
            PMTStream.FMT,
            self.streamtype,
            self.elementary_pid + 0xE000,
            len(self.elementary_stream_descriptors) + 0xF000,
        )
        return hdr + self.elementary_stream_descriptors

    def __len__(self) -> int:
        return struct.calcsize(PMTStream.FMT) + len(self.elementary_stream_descriptors)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, PMTStream):
            return False
        if (
            self.streamtype != value.streamtype
            or self.elementary_pid != value.elementary_pid
            or self.elementary_stream_descriptors != value.elementary_stream_descriptors
        ):
            return False
        return True

    def __repr__(self) -> str:
        r = f"Stream Type={self.streamtype:#04X} Elementary PID={self.elementary_pid:#06X} Descriptor len={len(self.elementary_stream_descriptors)}"
        return r


def crc32mpeg2(msg):
    """CRC poly used in the MPEG stream"""
    crc = 0xFFFFFFFF
    for b in msg:
        crc ^= b << 24
        for _ in range(8):
            crc = (crc << 1) ^ 0x04C11DB7 if crc & 0x80000000 else crc << 1
    return crc & 0xFFFFFFFF


# PMT Header = 12
class MPEGPacketPMT(MPEGPacket):
    """Specific version of an MPEG Packet which contains the PMT (program map specific data)
    https://en.wikipedia.org/wiki/Program-specific_information#PMT_(Program_map_specific_data)

    Args:
        MPEGPacket (_type_): _description_

    Returns:
        _type_: _description_
    """

    FMT = ">BHHBBBHH"
    FMT_POINTER = ">B"
    HDR_LEN_NOT_INCL_IN_LEN = 3
    CRC_LEN = 4

    def __init__(self):
        super().__init__()
        self.tableid: int = 0x0
        self.syntax_indicator = 0x0
        self.program_number: int = 0x0
        self.version: int = 0x3
        self.current_next_indicator: int = 0x0
        self.section: int = 0
        self.last_section: int = 0
        self.pcr_pid: int = 0x0
        self.program_info_len: int = 0x0
        self.streams: typing.List[PMTStream] = []
        self.descriptor_tags: typing.List[DescriptorTag] = []
        self._crc = None

    def unpack(self, buf: bytes) -> bool:
        """Unpack the buffer as a PMT packet

        Args:
            buf (bytes): _description_

        Returns:
            bool: CRC as expected
        """
        super().unpack(buf)
        (_pointer,) = struct.unpack_from(MPEGPacketPMT.FMT_POINTER, self.payload)
        logger.debug(f"Pointer={_pointer}")
        (self.tableid, _pmt, self.program_number, _ver, self.section, self.last_section, _pcr, _pil) = (
            struct.unpack_from(MPEGPacketPMT.FMT, self.payload, struct.calcsize(MPEGPacketPMT.FMT_POINTER) + _pointer)
        )
        self.syntax_indicator = _pmt >> 15
        _len = _pmt & 0xFFF
        self.version = (_ver >> 1) & 0x1F
        self.current_next_indicator = _ver & 0x1
        self.pcr_pid = _pcr & 0x1FFF
        self.program_info_len = _pil & 0xFFF
        _offset = struct.calcsize(MPEGPacketPMT.FMT) + struct.calcsize(MPEGPacketPMT.FMT_POINTER) + _pointer
        if self.program_info_len > 0:
            descriptor_tag_buffer = self.payload[_offset : _offset + self.program_info_len]
            while len(descriptor_tag_buffer) > 0:
                dt = DescriptorTag()
                descriptor_tag_buffer = dt.unpack(descriptor_tag_buffer)
                self.descriptor_tags.append(dt)
            _offset += self.program_info_len

        # Offset to the start of the streams

        # Length of the stream payload
        stream_len = (
            _len - struct.calcsize(MPEGPacketPMT.FMT) + MPEGPacketPMT.HDR_LEN_NOT_INCL_IN_LEN - self.program_info_len
        )
        stream_buf = self.payload[_offset : (stream_len + _offset)]
        # The buffer over which to calculate the CRC
        crc_buffer = self.payload[
            (_pointer + struct.calcsize(MPEGPacketPMT.FMT_POINTER)) : (_offset + stream_len - MPEGPacketPMT.CRC_LEN)
        ]
        exp_crc = crc32mpeg2(crc_buffer)
        logger.debug(f"crc_len={len(crc_buffer)} :{crc_buffer[0]} : {crc_buffer[-1]}")
        # extract all the streams
        while len(stream_buf) > MPEGPacketPMT.CRC_LEN:
            stream = PMTStream()
            stream_buf = stream.unpack(stream_buf)
            self.streams.append(stream)
        # All that is left is the CRC
        (self._crc,) = struct.unpack(">I", stream_buf)
        # Check it
        if self._crc != exp_crc:
            logger.error(f"CRC Does not match in PMT packet. Act={self._crc:#0X} exp={exp_crc:#0X}")
            return False
        else:
            return True

    def pack(self) -> bytes:
        """Convert MPEGPacketPMT into bytes

        Returns:
            bytes: _description_
        """
        _len = struct.calcsize(MPEGPacketPMT.FMT) - MPEGPacketPMT.HDR_LEN_NOT_INCL_IN_LEN + MPEGPacketPMT.CRC_LEN
        for s in self.streams:
            _len += len(s)
        self.program_info_len = 0
        for _t in self.descriptor_tags:
            _len += len(_t)
            self.program_info_len += len(_t)
        _buf = struct.pack(MPEGPacketPMT.FMT_POINTER, 0x0)
        _buf += struct.pack(
            MPEGPacketPMT.FMT,
            self.tableid,
            (self.syntax_indicator << 15) + (0x3 << 12) + _len,
            self.program_number,
            (0x3 << 6) + (self.version << 1) + self.current_next_indicator,
            self.section,
            self.last_section,
            (0x7 << 13) + self.pcr_pid,
            (0xF << 12) + self.program_info_len,
        )
        for _t in self.descriptor_tags:
            _buf += _t.pack()
        for s in self.streams:
            _buf += s.pack()
        _crc = crc32mpeg2(_buf[struct.calcsize(MPEGPacketPMT.FMT_POINTER) :])
        self.payload = _buf + struct.pack(">L", _crc)

        return super().pack()

    def __repr__(self) -> str:
        r = super().__repr__()
        r += f"PMT: TableID={self.tableid:#04X} ProgramNumer={self.program_number:#06X} SectionNumber={self.section} LastSection={self.last_section} PCR={self.pcr_pid:#06X} PiL={self.program_info_len}\n"
        for s in self.streams:
            r += f"{repr(s)}\n"
        for _t in self.descriptor_tags:
            r += f"{repr(_t)}\n"
        return r

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MPEGPacketPMT):
            return False
        match_attr = [
            "sync",
            "pid",
            "transport_priority",
            "tei",
            "pusi",
            "continuitycounter",
            "tsc",
            "adaption_ctrl",
            "adaption_field",
            "tableid",
            "syntax_indicator",
            "program_number",
            "section",
            "last_section",
            "pcr_pid",
            "program_info_len",
            "streams",
            "descriptor_tags",
        ]
        for attr in match_attr:
            if getattr(self, attr) != getattr(__value, attr):
                logger.error(f"Attr={attr}, Self={getattr(self, attr)} Other={getattr(__value, attr)} ")
                return False
            return True



"""
.. module:: ParserAlignedPacket
    :platform: Unix, Windows
    :synopsis: Class to pack and unpack Parser Aligned payloads

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

import struct
import typing

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


class ParserAlignedBlock(object):
    """
    A class to handle a single Parser Block. Returns an object containing all the fields of a parser block

    """

    def __init__(self):
        self.error: bool = False  #: Error Flag
        self.errorcode: int = 0  #: Error code field
        self.quadbytes: int = 1  #: Number of quadbytes in the parser block
        self.messagecount: int = 0  #: Wrapping 8 bit message counter
        self.busid: int = 1  #: Bus ID on which the message was captured
        self.elapsedtime: int = 2  #: Time tag in nanoseconds offset from the iNetx timestamp
        self.payload: bytes = bytes()  #: Payload

        self.format = ">HBBL"
        self.headerlen = struct.calcsize(self.format)

    def unpack(self, buf: bytes):
        """
        Unpack a single parser block. Unsually called only from the ParserAlignedPacket unpack method.

        Returns the length of the parser block

        :param buf: The string buffer to unpack
        :type buf: str
        :rtype: int
        """

        # Unpack the fields
        (error_and_quad, self.messagecount, self.busid, self.elapsedtime) = struct.unpack_from(self.format, buf)
        # MSB bit is the error flag
        if (error_and_quad >> 15) == 1:
            self.error = True
        else:
            self.error = False

        # As error code and quad bytes are not byte aligned I have to do some shifting and masking
        self.errorcode = (error_and_quad >> 9) & 0x3F
        self.quadbytes = error_and_quad & 0x1FF
        payload_length_in_bytes = (self.quadbytes - 2) * 4

        if self.quadbytes < 2:
            raise ValueError("The quad bytes cannot be less than 2. Actual={}".format(self.quadbytes))

        # Raise an exception if the payload is not as expected
        if len(buf) < self.headerlen + payload_length_in_bytes:
            raise ValueError(
                "The length of the block buffer {} is smaller the expect length {}".format(
                    len(buf), self.headerlen + payload_length_in_bytes
                )
            )

        self.payload = buf[self.headerlen : self.quadbytes * 4]
        # Return the length of the block as a courtesy
        return self.quadbytes * 4

    def pack(self) -> bytes:
        """
        Convert a ParserAlignedBlock into a buffer
        :return:
        """
        if len(self.payload) % 4 != 0:
            raise Exception("Length of payload is not aligned to quadbytes (32b)")
        self.quadbytes = 2 + len(self.payload) // 4
        error_and_quad = (self.error << 15) + ((self.errorcode & 0x3F) << 9) + self.quadbytes
        return struct.pack(self.format, error_and_quad, self.messagecount, self.busid, self.elapsedtime) + self.payload

    def __repr__(self):
        return "QuadBytes={} Error={} ErrorCode={} BusID={} MessageCount={} ElapsedTime={}".format(
            self.quadbytes, self.error, self.errorcode, self.busid, self.messagecount, self.elapsedtime
        )

    def __len__(self):
        return len(self.payload) + struct.calcsize(self.format)

    def __eq__(self, other):
        if not isinstance(other, ParserAlignedBlock):
            return False
        for attr in ["quadbytes", "error", "errorcode", "busid", "messagecount", "elapsedtime", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class ParserAlignedPacket(object):
    """
    A class that handles iNetx parser aligned packets. Unpack a buffer to populate the field into a list of parserblocks

    Capture a UDP packet,unpack as iNetX whose payload is parser aligned

    >>> from base64 import b64decode
    >>> import AcraNetwork.iNetX as inetx
    >>> data = b64decode('EQAAAAAAAAEAAAAAAAAAQAAAAAAAAAAAAAAAAAADAAAAAAAAQJ+3iAADAQIAAAAAekKvKAADAgYAACcQ7Z6oWA==')
    >>> i = inetx.iNetX()
    >>> i.unpack(data)
    True
    >>> p = ParserAlignedPacket()
    >>> p.unpack(i.payload)
    True
    >>> print(p[0])
    QuadBytes=3 Error=False ErrorCode=0 BusID=0 MessageCount=0 ElapsedTime=0


    :type parserblocks: list[ParserAlignedBlock]
    :type numberofblocks: int
    """

    def __init__(self):
        self.parserblocks: typing.List[ParserAlignedBlock] = []  #: List of ParserAlignedBlock
        self.numberofblocks: int = 0  #: The number of ParserAlignedBlock

    def unpack(self, buf: bytes):
        """
        Pass a buffer containing all the Parser alignd payload and this method will
        unpack all the fields. Returns a list of parserblocks

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """
        fullbufferlen = len(buf)
        bufferparsed = 0
        while bufferparsed < fullbufferlen:
            block = ParserAlignedBlock()
            # unpack and add the amount unpacked to the running total
            bufferparsed += block.unpack(buf[bufferparsed:])
            self.parserblocks.append(block)
        return True

    def pack(self):
        """Convert a ParserPacket to a buffer"""

        buf = bytes()
        for b in self.parserblocks:
            buf += b.pack()

        return buf

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parserblocks):
            _block = self.parserblocks[self._index]
            self._index += 1
            return _block
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parserblocks)

    def __getitem__(self, key):
        return self.parserblocks[key]

    def __repr__(self):
        rep = ""
        for idx, b in enumerate(self.parserblocks):
            rep += "Block {}: {}\n".format(idx, repr(b))
        return rep

    def __eq__(self, other):
        if not isinstance(other, ParserAlignedPacket):
            return False
        if len(other) != len(self):
            return False
        blk_cnt = len(other)
        for idx in range(blk_cnt):
            if other.parserblocks[idx] != self.parserblocks[idx]:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class ARINC429(object):
    """This is not working yet. Don't use it"""

    MESSAGE_LEN = 4  # 4 bytes
    LABEL_REVERSE = [
        0x00,
        0x80,
        0x40,
        0xC0,
        0x20,
        0xA0,
        0x60,
        0xE0,  # 0 -   8
        0x10,
        0x90,
        0x50,
        0xD0,
        0x30,
        0xB0,
        0x70,
        0xF0,  # 9 -  15
        0x08,
        0x88,
        0x48,
        0xC8,
        0x28,
        0xA8,
        0x68,
        0xE8,  # 16 -  23
        0x18,
        0x98,
        0x58,
        0xD8,
        0x38,
        0xB8,
        0x78,
        0xF8,  # 24 -  31
        0x04,
        0x84,
        0x44,
        0xC4,
        0x24,
        0xA4,
        0x64,
        0xE4,  # 32 -  39
        0x14,
        0x94,
        0x54,
        0xD4,
        0x34,
        0xB4,
        0x74,
        0xF4,  # 40 -  47
        0x0C,
        0x8C,
        0x4C,
        0xCC,
        0x2C,
        0xAC,
        0x6C,
        0xEC,  # 48 -  55
        0x1C,
        0x9C,
        0x5C,
        0xDC,
        0x3C,
        0xBC,
        0x7C,
        0xFC,  # 56 -  63
        0x02,
        0x82,
        0x42,
        0xC2,
        0x22,
        0xA2,
        0x62,
        0xE2,  # 64 -  71
        0x12,
        0x92,
        0x52,
        0xD2,
        0x32,
        0xB2,
        0x72,
        0xF2,  # 72 -  79
        0x0A,
        0x8A,
        0x4A,
        0xCA,
        0x2A,
        0xAA,
        0x6A,
        0xEA,  # 80 -  87
        0x1A,
        0x9A,
        0x5A,
        0xDA,
        0x3A,
        0xBA,
        0x7A,
        0xFA,  # 88 -  95
        0x06,
        0x86,
        0x46,
        0xC6,
        0x26,
        0xA6,
        0x66,
        0xE6,  # 96 - 103
        0x16,
        0x96,
        0x56,
        0xD6,
        0x36,
        0xB6,
        0x76,
        0xF6,  # 104 - 111
        0x0E,
        0x8E,
        0x4E,
        0xCE,
        0x2E,
        0xAE,
        0x6E,
        0xEE,  # 112 - 119
        0x1E,
        0x9E,
        0x5E,
        0xDE,
        0x3E,
        0xBE,
        0x7E,
        0xFE,  # 120 - 127
        0x01,
        0x81,
        0x41,
        0xC1,
        0x21,
        0xA1,
        0x61,
        0xE1,  # 128 - 135
        0x11,
        0x91,
        0x51,
        0xD1,
        0x31,
        0xB1,
        0x71,
        0xF1,  # 136 - 143
        0x09,
        0x89,
        0x49,
        0xC9,
        0x29,
        0xA9,
        0x69,
        0xE9,  # 144 - 151
        0x19,
        0x99,
        0x59,
        0xD9,
        0x39,
        0xB9,
        0x79,
        0xF9,  # 152 - 159
        0x05,
        0x85,
        0x45,
        0xC5,
        0x25,
        0xA5,
        0x65,
        0xE5,  # 160 - 167
        0x15,
        0x95,
        0x55,
        0xD5,
        0x35,
        0xB5,
        0x75,
        0xF5,  # 168 - 175
        0x0D,
        0x8D,
        0x4D,
        0xCD,
        0x2D,
        0xAD,
        0x6D,
        0xED,  # 176 - 183
        0x1D,
        0x9D,
        0x5D,
        0xDD,
        0x3D,
        0xBD,
        0x7D,
        0xFD,  # 184 - 191
        0x03,
        0x83,
        0x43,
        0xC3,
        0x23,
        0xA3,
        0x63,
        0xE3,  # 192 - 199
        0x13,
        0x93,
        0x53,
        0xD3,
        0x33,
        0xB3,
        0x73,
        0xF3,  # 200 - 207
        0x0B,
        0x8B,
        0x4B,
        0xCB,
        0x2B,
        0xAB,
        0x6B,
        0xEB,  # 208 - 215
        0x1B,
        0x9B,
        0x5B,
        0xDB,
        0x3B,
        0xBB,
        0x7B,
        0xFB,  # 216 - 223
        0x07,
        0x87,
        0x47,
        0xC7,
        0x27,
        0xA7,
        0x67,
        0xE7,  # 224 - 231
        0x17,
        0x97,
        0x57,
        0xD7,
        0x37,
        0xB7,
        0x77,
        0xF7,  # 232 - 239
        0x0F,
        0x8F,
        0x4F,
        0xCF,
        0x2F,
        0xAF,
        0x6F,
        0xEF,  # 240 - 247
        0x1F,
        0x9F,
        0x5F,
        0xDF,
        0x3F,
        0xBF,
        0x7F,
        0xFF,  # 248 - 255
        0x00,
    ]

    def __init__(self):
        self.parity = None
        self.ssm = None
        self.data = None
        self.sdi = None
        self.label = None

    def unpack(self, buf):
        if len(buf) != ARINC429.MESSAGE_LEN:
            raise ValueError("Buffer is not the correct length for an ARINC429 message")
        (byte1, byte2, byte3, byte4) = struct.unpack("BBBB", buf)
        self.parity = byte1 / 128
        self.ssm = (byte1 / 32) % 4
        self.data = ((byte1 % 32) * 256 + byte2) * 64 + (byte3 / 4)
        self.sdi = byte3 % 4
        self.label = ARINC429.LABEL_REVERSE[byte4]



"""
.. module:: iNetX
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct iNetx Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct


class iNetX(object):
    """
    Class to pack and unpack iNetX payloads. iNet-X is an open payload format for use
    in FTI networks. It is usually transmitted in a UDP packet containing parameter data
    acquired from sensors and buses

    Capture a UDP packet and unpack the payload as an iNetX packet

    Here's an example where you take some bytes, decode it as an iNetX packet, then build the same
    packet from scratch and compare the two


    >>> from base64 import b64decode
    >>> data = b64decode('EQAAAAAAANwAAAABAAAAHgAAAAEAAAABAAAAAAUA')
    >>> pkt = iNetX()
    >>> pkt.unpack(data)
    True
    >>> print(f"{pkt.streamid:#0X}")
    0XDC
    >>> i = iNetX()
    >>> i.sequence = 1
    >>> i.pif = 0
    >>> i.streamid = 0xDC
    >>> i.setPacketTime(1, 1)
    True
    >>> i.payload = struct.pack("H", 0x5)
    >>> i == pkt
    True

    """

    DEF_CONTROL_WORD = 0x11000000  #: (Object Constant) The default iNetX control word.
    INETX_HEADER_FORMAT = ">LLLLLLL"
    INETX_HEADER_LENGTH = struct.calcsize(INETX_HEADER_FORMAT)
    REQ_ATTR = ("inetxcontrol", "streamid", "sequence", "ptptimeseconds", "ptptimenanoseconds", "pif", "payload")

    def __init__(self, buf: bytes = None):
        """Creator method for an iNetX class

        Args:
            buf (bytes, optional): Optionally unpack the byte buffer into an iNetX object. Defaults to None.
        """
        self.inetxcontrol: int = iNetX.DEF_CONTROL_WORD  #: Control Word Defaults to 0x11000000
        self.streamid: int = 0  #: Stream ID. Typically to identify a unique packet in an FTI network. 4 bytes in size
        self.sequence: int = 0  #: Unique rollover counter per stream ID.Rolls over at 2^64
        self.packetlen: int = 0  #: Packet Length in bytes
        self.ptptimeseconds: int = 0  #: Timestamp of first parameter in the packet. EPOCH time
        self.ptptimenanoseconds: int = 0  #: Nanaosecond timestamp
        self.pif: int = 0  #: Payload Information Field
        self.payload: bytes = bytes()  #: Payload

        self._packetStrut = struct.Struct(iNetX.INETX_HEADER_FORMAT)
        if buf is not None:
            self.unpack(buf)

    def pack(self) -> bytes:
        """
        Pack the packet into a binary format and return as a string

        :rtype: bytes
        """

        for attr in iNetX.REQ_ATTR:
            if getattr(self, attr) is None:
                raise ValueError("Require {} is not defined".format(attr))

        self.packetlen = len(self.payload) + iNetX.INETX_HEADER_LENGTH
        packetvalues = (
            self.inetxcontrol,
            self.streamid,
            self.sequence,
            self.packetlen,
            self.ptptimeseconds,
            self.ptptimenanoseconds,
            self.pif,
        )
        packet = self._packetStrut.pack(*packetvalues) + self.payload
        return packet

    def unpack(self, buf: bytes) -> bool:
        """
        Unpack a raw byte stream to an iNetX object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        if len(buf) < iNetX.INETX_HEADER_LENGTH:
            raise ValueError("Buffer is too short to be an iNetX packet")

        (
            self.inetxcontrol,
            self.streamid,
            self.sequence,
            self.packetlen,
            self.ptptimeseconds,
            self.ptptimenanoseconds,
            self.pif,
        ) = self._packetStrut.unpack_from(buf)

        if self.packetlen != len(buf):
            raise ValueError(
                "Length of buffer 0x{:X} does not match length field 0x{:X}".format(len(buf), self.packetlen)
            )

        self.payload = buf[iNetX.INETX_HEADER_LENGTH :]

        return True

    def setPacketTime(self, utctimestamp, nanoseconds=0):
        """
        Set the packet timestamp

        :param timestamp: The timestamp in seconds since 1 Jan 1970
        :type timestamp: int
        :type nanoseconds: Nanoseconds past the current time
        :type nanoseconds: int
        """

        self.ptptimeseconds = utctimestamp
        self.ptptimenanoseconds = nanoseconds

        return True

    def __repr__(self):
        return "STREAMID={:#0X} SEQ={} LEN={} PTPS={} PTPNS={}".format(
            self.streamid, self.sequence, self.packetlen, self.ptptimeseconds, self.ptptimenanoseconds
        )

    def __eq__(self, other):
        if not isinstance(other, iNetX):
            return False

        for attr in iNetX.REQ_ATTR:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __len__(self):
        return len(self.pack())



#!/usr/bin/env python3

# -*- coding: utf-8 -*-


import AcraNetwork.IRIG106.Chapter11 as ch11
from AcraNetwork.IRIG106.Chapter10 import FileParser
import AcraNetwork.IRIG106.Chapter10.Chapter10UDP as ch10udp
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as eth
import argparse
import sys


def create_parser():
    # ----------------------------------
    # Setup the command line parser
    # ----------------------------------
    parser = argparse.ArgumentParser(description="Covert a chapter 10 file to a pcap")
    parser.add_argument("--pcap", required=True, help="The output pcap file")
    parser.add_argument("--ch10", required=True, help="The input chapter 10 file")
    parser.add_argument("--tmats", required=False, default=None, help="Optional TMATS output file")
    return parser


def encapsulate_udppayload_in_eth(udp_payload: bytes):
    """
    Encapsulte the udp payload in an Ethernet packet
    """
    ethpkt = eth.Ethernet()
    ethpkt.dstmac = 0x01005E000001
    ethpkt.srcmac = 0x000C4DAC7AAA
    ethpkt.type = eth.Ethernet.TYPE_IP
    #
    ippkt = eth.IP()
    ippkt.dstip = "235.0.0.2"
    ippkt.srcip = "127.0.0.1"
    # Stick a UDP packet in the payload
    udppkt = eth.UDP()
    udppkt.dstport = 51001
    udppkt.srcport = 51001

    udppkt.payload = udp_payload
    # packet the udp packet into the ethernet payload
    ippkt.payload = udppkt.pack()
    ethpkt.payload = ippkt.pack()
    return ethpkt.pack()


def main(args):
    pf = pcap.Pcap(args.pcap, mode="w")
    fp = FileParser.FileParser(args.ch10)
    if args.tmats is not None:
        tf = open(args.tmats, mode="wb")

    idx = 0
    with fp as ch10file:
        for idx, _payload in enumerate(ch10file):
            pkt = ch11.Chapter11()
            try:
                pkt.unpack(_payload)
            except Exception as e:
                print(f"Failed to unpack. err={e}")
            else:
                if args.tmats is not None and idx == 0:
                    tf.write(pkt.pack())
                    tf.close()
                pr = pcap.PcapRecord()
                pr.set_current_time()
                udp = ch10udp.Chapter10UDP()
                udp.format = 3
                udp.sourceid_len = 0
                udp.sequence = idx
                udp.offset_pkt_start = 0
                udp.payload = pkt.pack()
                pr.payload = encapsulate_udppayload_in_eth(udp.pack())
                pf.write(pr)

    pf.close()
    print(f"Create a pcap with {idx} records")


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    ret = main(args)
    sys.exit(ret)
